#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import os
import shutil
import subprocess

import pytest
from sos import execute_workflow
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env


def test_progress_bar():
    # progress bar with nested workflow
    execute_workflow(
        '''
import time
time.sleep(0)
[sub_1]
[sub_2]
[sub_3]
[sub_4]
[a_1]
[a_2]
[a_3]
sos_run('sub')
[a_4]
[a_5]
''',
        workflow='a')


def test_nested_workflow(temp_factory, clear_now_and_after):
    '''Test the creation and execution of combined workfow'''
    clear_now_and_after('a.done')
    temp_factory('a.txt', 'b.txt', 'b.begin')
    execute_workflow(
        '''
        if 'executed' not in locals():
            executed = []
        if 'inputs' not in locals():
            inputs = []

        [a_1: shared=['executed', 'inputs']]
        executed.append(step_name)
        inputs.append(_input)

        [a_2: shared=['executed', 'inputs']]
        executed.append(step_name)
        inputs.append(_input)

        [a_3: shared=['executed', 'inputs']]
        executed.append(step_name)
        inputs.append(_input)

        [a_4: shared=['executed', 'inputs']]
        executed.append(step_name)
        output: 'a.done'
        inputs.append(_input)
        run: expand=True
        touch {_output}

        [b_1: shared=['executed', 'inputs']]
        executed.append(step_name)
        input: 'b.begin'
        inputs.append(_input)

        [b_2: shared=['executed', 'inputs']]
        executed.append(step_name)
        inputs.append(_input)

        [b_3: shared=['executed', 'inputs']]
        executed.append(step_name)
        inputs.append(_input)

        [b_4: shared=['executed', 'inputs']]
        executed.append(step_name)
        output: 'b.txt'
        inputs.append(_input)

        [c: shared=['executed', 'inputs']]
        executed.append(step_name)
        input: 'a.txt'
        output: 'b.txt'
        inputs.append(_input)
        sos_run('a+b', shared=['executed', 'inputs'])
        ''',
        workflow='c',
        options={'sig_mode': 'ignore'})

    # order of execution is not guaranteed
    assert sorted(env.sos_dict['executed']) == sorted(
        ['c', 'a_1', 'a_2', 'a_3', 'a_4', 'b_1', 'b_2', 'b_3', 'b_4'])
    env.sos_dict.pop('executed', None)


def test_looped_nested_workflow(temp_factory, clear_now_and_after):
    # step will be looped
    clear_now_and_after('a.txt.a1', 'a.txt.a1.a2', 'b.txt.a1', 'b.txt.a1.a2')
    temp_factory('a.txt', 'b.txt')

    execute_workflow(
        '''
        if 'executed' not in locals():
            executed = []
        if 'inputs' not in locals():
            inputs = []

        [a_1:shared=['executed', 'inputs']]
        executed.append(step_name)
        output: _input[0] + '.a1'
        inputs.append(_input)
        run: expand=True
        touch {_output}

        [a_2:shared=['executed', 'inputs']]
        executed.append(step_name)
        output: _input[0] + '.a2'
        inputs.append(_input)
        run: expand=True
        touch {_output}

        [c:shared=['executed', 'inputs']]
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        inputs.append(_input)
        sos_run('a', shared=['executed', 'inputs'])
        ''',
        workflow='c')
    assert env.sos_dict['executed'] == ['c', 'a_1', 'a_2', 'a_1', 'a_2']


def test_single_looped_nested_workflow(temp_factory, clear_now_and_after):
    temp_factory('a.txt', 'b.txt')
    env.sos_dict.pop('executed', None)

    execute_workflow(
        '''
        if 'executed' not in locals():
            executed = []
        [a_1:shared='executed']
        executed.append(step_name)
        [a_2:shared='executed']
        executed.append(step_name)
        [c_0:shared='executed']
        executed.append(step_name)
        [c_1:shared='executed']
        depends: sos_variable('executed')
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        sos_run('a:2', shared='executed')
        ''',
        workflow='c')
    assert env.sos_dict['executed'] == ['c_0', 'c_1', 'a_2', 'a_2']
    env.sos_dict.pop('executed', None)
    # allow specifying a single step
    # step will be looped
    execute_workflow(
        '''
        if 'executed' not in locals():
            executed = []

        [a_1:shared='executed']
        executed.append(step_name)

        [a_2:shared='executed']
        executed.append(step_name)

        [c_0:shared='executed']
        executed.append(step_name)

        [c_1:shared='executed']
        depends: sos_variable('executed')
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        sos_run('a:2', shared='executed')
        ''',
        workflow='c')

    assert env.sos_dict['executed'] == ['c_0', 'c_1', 'a_2', 'a_2']


def test_recursive_nested_workflow(temp_factory):
    # recursive subworkflow not allowed
    temp_factory('a.txt', 'b.txt')
    with pytest.raises(Exception):
        execute_workflow(
            '''
            if 'executed' not in locals():
                executed = []
            [a_1:shared='executed']
            executed.append(step_name)
            [a_2:shared='executed']
            executed.append(step_name)
            [c_0:shared='executed']
            executed.append(step_name)
            [c_1:shared='executed']
            executed.append(step_name)
            input: 'a.txt', 'b.txt', group_by='single'
            sos_run('a_2+c', shared='executed')
            ''',
            workflow='c')
    #
    env.sos_dict.pop('executed', None)
    # nested subworkflow is allowed
    execute_workflow(
        '''
        if 'executed' not in locals():
            executed = []
        [a_1:shared='executed']
        executed.append(step_name)
        [a_2:shared='executed']
        executed.append(step_name)
        [a_3:shared='executed']
        executed.append(step_name)
        [b_1:shared='executed']
        executed.append(step_name)
        [b_2:shared='executed']
        executed.append(step_name)
        sos_run('a:1-2', shared='executed')
        [c_0:shared='executed']
        executed.append(step_name)
        [c_1:shared='executed']
        depends: sos_variable('executed')
        executed.append(step_name)
        input: 'a.txt'
        sos_run('a+b', shared='executed')
        ''',
        workflow='c')

    assert env.sos_dict['executed'] == [
        'c_0', 'c_1', 'a_1', 'a_2', 'a_3', 'b_1', 'b_2', 'a_1', 'a_2'
    ]

    env.sos_dict.pop('executed', None)


def test_subworkflow_with_options(temp_factory, clear_now_and_after):
    # nested subworkflow with step option and others
    temp_factory('a.txt', 'b.txt')
    clear_now_and_after('a.done')
    script = '''
        if 'executed' not in locals():
            executed = []
        [a_1:shared='executed']
        executed.append(step_name)

        [a_2:shared='executed']
        executed.append(step_name)

        [a_3:shared='executed']
        executed.append(step_name)

        [b:shared='executed']
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        sos_run('a:3+a:1', shared='executed')

        [d:shared='executed']
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        sos_run('a:2', shared='executed')

        [e2_2:shared='executed']
        executed.append(step_name)
        input: 'a.txt', 'b.txt', group_by='single'
        '''
    execute_workflow(script, workflow='b')
    assert env.sos_dict['executed'] == ['b', 'a_3', 'a_1', 'a_3', 'a_1']

    env.sos_dict.pop('executed', None)

    execute_workflow(script, workflow='d')
    assert env.sos_dict['executed'] == ['d', 'a_2', 'a_2']
    env.sos_dict.pop('executed', None)

    execute_workflow(script, workflow='e2')
    assert env.sos_dict['executed'] == ['e2_2']


def test_dynamic_nested_workflow():
    '''Test nested workflow controlled by command line option'''
    script = '''
        if 'executed' not in locals():
            executed = []

        parameter: wf='a'

        [a_1:shared='executed']
        executed.append(step_name)
        [a_2:shared='executed']
        executed.append(step_name)
        [a_3:shared='executed']
        executed.append(step_name)
        [b_1:shared='executed']
        executed.append(step_name)
        [b_2:shared='executed']
        executed.append(step_name)
        [b_3:shared='executed']
        executed.append(step_name)

        [default:shared='executed']
        executed.append(step_name)
        sos_run(wf, shared='executed')
        '''
    execute_workflow(script, args=['--wf', 'b'])
    assert env.sos_dict['executed'] == ['default', 'b_1', 'b_2', 'b_3']
    #
    env.sos_dict.pop('executed', None)
    execute_workflow(script, args=['--wf', 'a'])
    assert env.sos_dict['executed'] == ['default', 'a_1', 'a_2', 'a_3']


def test_sos_run(clear_now_and_after):
    '''Test action sos_run with keyword parameters'''
    clear_now_and_after('0.txt', '1.txt')
    execute_workflow(
        r'''
        [A]
        parameter: num=5

        run: expand=True
            touch {num}.txt

        [batch]
        for k in range(2):
            sos_run('A', num=k)
        ''',
        workflow='batch')

    for f in ['0.txt', '1.txt']:
        assert file_target(f).target_exists()
    #
    clear_now_and_after('0.txt', '1.txt', '5.txt')

    execute_workflow(
        r'''
        [A]
        parameter: num=5
        run: expand=True
            touch {num}.txt

        [batch]
        for num in range(2):
            sos_run('A')
        ''',
        workflow='batch')
    for f in ['0.txt', '1.txt']:
        assert not file_target(f).target_exists()

    assert file_target('5.txt').target_exists()

    clear_now_and_after('10.txt', '11.txt')
    #
    # test parameter shared to send and return vars
    #
    execute_workflow(
        r'''
        [A: shared='k']
        k += 10

        [batch]
        for k in range(2):
            sos_run('A', shared='k')
            run(f"touch {k}.txt")
        ''',
        workflow='batch')
    for f in ['10.txt', '11.txt']:
        assert file_target(f).target_exists()


def test_da_gof_dynamic_nested_workflow(clear_now_and_after):
    #
    # Because we are not sure which workflows would be executed
    # until run time, the DAG should not contain nested workflow
    # until runtime.
    #
    clear_now_and_after('B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt',
                        'B2.txt.p')
    #
    #  A1 <- P <- B
    #  A1 <- P <- B
    #  A2
    #
    #  ALL calls A and B with parameter
    #
    execute_workflow(
        '''
        [A_1]
        parameter: num = 2
        input: f"B{num}.txt.p"

        [B: provides='B{num}.txt']
        run: expand=True
        touch 'B{num[0]}.txt'

        [P: provides='{filename}.p']
        input: filename
        run: expand=True
        touch {_output}

        [ALL]

        for i in range(3):
            sos_run('A', num=i)
    ''',
        workflow='ALL')
    for f in ['B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p']:
        assert file_target(f).target_exists()


def test_outcome_oriented_nested_workflow(clear_now_and_after):
    '''test nested workflow triggered by targets'''
    clear_now_and_after('test_15.txt')

    execute_workflow('''
[A: provides='test_{idx}.txt']
_output.touch()

[default]
sos_run(targets='test_15.txt')
    ''')
    assert os.path.isfile('test_15.txt')


def test_passing_vars_to_nested_workflow():
    '''Test if variables can be passed to nested workflows'''
    execute_workflow(r"""
        import time
        import random

        [nested]
        parameter: nested=True
        parameter: seed=1
        print(f'I am nested {nested} with seed {seed}')

        [0]
        reps = range(5)
        input: for_each='reps'
        import random
        nested = _reps
        seed = random.randint(1, 1000)
        print(f'Passing {seed} to {nested}')
        sos_run('nested', nested=nested, seed=seed)
    """)


def test_user_defined_func():
    '''Test the use of user-defined functions in SoS script'''
    execute_workflow(
        r"""
        def myfunc():
            return 'a'

        [1: shared={'test':'_output'}]
        output: myfunc()

        myfunc()

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['test'] == ['a']
    # User defined function should also work under nested workflows
    # This is difficult because the 'local namespace' is usually
    # not seen inside function definition. The solution now is to
    # use a single workspace.
    execute_workflow(
        r"""
        def myfunc():
            # test if builtin functions (sum and range) can be used here.
            return 'a' + str(sum(range(10)))

        [1: shared={'test':'_output'}]
        output: [myfunc() for i in range(10)][0]

        myfunc()

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['test'] == ['a45']


def test_config_file_of_nested_workflow(config_factory):
    '''Test passing of configurationg to nested workflow'''
    cfg = config_factory("""{'1':'hi'}""")

    execute_workflow(
        '''
[test_1]
parameter: key = None
print(CONFIG[key])

[default_1]
sos_run('test:1', key = '1')
''',
        options={'config_file': cfg})


def test_error_from_subworkflow():
    '''Test if error from subworkflow is passed to master (#396)'''
    with pytest.raises(Exception):
        execute_workflow('''
            [test_1]
            R:
            set.seed(xxx)

            [default]
            sos_run('test')
            ''')


def test_fun_def(temp_factory):
    '''Test defintion of function that can be used by other steps'''
    temp_factory('aa.txt', 'ab.txt')
    # in nested workflow?
    execute_workflow(
        r"""
        def myfunc(a):
            return ['a' + x for x in a]

        [mse: shared={'test':'_output'}]
        input: myfunc(['a.txt', 'b.txt'])

        [1]
        sos_run('mse')
        """,
        options={'run)mode': 'dryrun'})
    #
    # Names defined in subworkflow is not returned to the master dict
    assert 'test' not in env.sos_dict


def test_search_path(clear_now_and_after):
    '''Test if any action should exit in five seconds in dryrun mode'''
    clear_now_and_after('crazy_path', 'test.yml')

    sos_config_file = os.path.join(
        os.path.expanduser('~'), '.sos', 'config.yml')
    shutil.copy(sos_config_file, 'test.yml')
    #
    subprocess.call(
        'sos config --set sos_path {0}/crazy_path {0}/crazy_path/more_crazy/'
        .format(os.getcwd()),
        shell=True)
    #
    if not os.path.isdir('crazy_path'):
        os.mkdir('crazy_path')
        os.mkdir(os.path.join('crazy_path', 'more_crazy'))
    with open(os.path.join('crazy_path', 'crazy_master.sos'), 'w') as crazy:
        crazy.write('''
[0]
sos_run('cc', source='crazy_slave.sos')

''')
    with open(os.path.join('crazy_path', 'more_crazy', 'crazy_slave.sos'),
              'w') as crazy:
        crazy.write('''
[cc_0]
print('hay, I am crazy')
''')

    script = SoS_Script(filename='crazy_master.sos')
    script.workflow()
    #
    os.remove(sos_config_file)
    shutil.copy('test.yml', sos_config_file)


def test_nested_workdir(clear_now_and_after, purge_tasks):
    '''Test nested runtime option for work directory'''
    clear_now_and_after('tmp', 'tmp1')
    execute_workflow(
        '''
        [step]
        task: workdir='tmp1'
        run:
            touch 'a.txt'

        [default]
        sos_run('step', workdir='tmp')
        ''',
        options={'default_queue': 'localhost'})

    assert os.path.isfile('tmp1/a.txt')


def test_failure_of_nested_workflow(clear_now_and_after):
    '''Test failure of nested workflow #838'''
    clear_now_and_after('a.txt')
    with pytest.raises(Exception):
        execute_workflow('''
            [something]
            input: 'a.txt'

            [default]
            sos_run('something')
            ''')


def test_nested_from_another_file(clear_now_and_after):
    '''Test nested runtime option for work directory'''
    clear_now_and_after('a.txt', 'another.sos')
    with open('another.sos', 'w') as another:
        another.write('''
[whatever]
run:
touch 'a.txt'

''')
    execute_workflow('''
        [default]
        sos_run('whatever', source='another.sos')
        ''')
    assert os.path.isfile(
        'a.txt'
    ), 'a.txt should have been created by nested workflow from another file'


def test_concurrent_sub_workflow():
    '''Test concurrent subworkflow sos_run '''
    import time
    st = time.time()

    execute_workflow(
        '''
        [A]
        parameter: idx=0
        import time
        time.sleep(5)

        [default]
        input: for_each=dict(i=range(6))
        sos_run('A', idx=i)
        ''',
        options={'worker_procs': ['8']})
    assert time.time() - st < 30


def test_sos_multi_workflow():
    '''Test multiple workflows in sos_run '''
    import time
    st = time.time()

    execute_workflow(
        '''
        [B]
        parameter: idx=2
        import time
        time.sleep(idx)

        [A]
        import time
        time.sleep(idx)

        [default]
        input: for_each=dict(i=range(4))
        sos_run(['A', 'B'], idx=i)
    ''',
        options={'worker_procs': ['8']})
    assert time.time() - st < 20


def test_pass_of_args():
    '''Test passing of arguments through sos_run #1164'''
    with pytest.raises(Exception):
        execute_workflow('''
            parameter: b=2
            [A]
            parameter: c=1
            print(c)

            [default]
            sos_run('A', a=2)
        ''')
    #
    execute_workflow('''
        parameter: b=2
        [A]
        parameter: c=1
        print(c)

        [default]
        sos_run('A', b=2)
        ''')
    #
    execute_workflow('''
        parameter: b=2
        [A]
        parameter: c=1
        print(c)

        [default]
        sos_run('A', c=2)
    ''')


def test_nested_dynamic_depends(clear_now_and_after):
    '''Test the execution of nested workflow with dynamic depends'''
    clear_now_and_after('B30.txt', 'B30.txt.p', 'B0.txt.p', 'B0.txt')
    execute_workflow('''
        [A_1]
        parameter: num = 20
        input: f"B{num}.txt"
        depends: _input + '.p'

        [B: provides='B{num}.txt']
        run: expand=True
            touch {_output}

        [P: provides='{filename}.p']
        input: filename
        run: expand=True
            touch {_output}

        [default]
        sos_run('A', num=0)
    ''')


def test_neste_with_both_workflow_and_targets(clear_now_and_after):
    '''Test nested workflow with both workflow and targets'''
    clear_now_and_after('A_out.txt', 'B_out.txt')
    execute_workflow(r'''
        [A_20]
        output: 'A_out.txt'
        _output.touch()

        [B]
        output: 'B_out.txt'
        _output.touch()

        [default]
        sos_run('A', targets='B_out.txt')
        ''')
    assert os.path.isfile('A_out.txt')
    assert os.path.isfile('B_out.txt')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os

import pytest
from sos import execute_workflow
# if the test is imported under sos/test, test interacive executor
from sos.workflow_executor import Base_Executor


def test_plain_target(clear_now_and_after):
    '''Test sos run -t filename'''
    clear_now_and_after('t_a.txt')
    execute_workflow(
        '''
        [A]
        output: 't_a.txt'
        _output.touch()
        ''', targets=['t_a.txt'])
    assert os.path.isfile('t_a.txt')


def test_target_in_named(clear_now_and_after):
    '''Test sos run -t filename'''
    clear_now_and_after('t_na.txt')
    execute_workflow(
        '''
        [A]
        output: res='t_na.txt'
        _output.touch()
        ''',
                targets=['t_na.txt'])
    assert os.path.isfile('t_na.txt')


def test_named_output_as_target(clear_now_and_after):
    '''Test sos run -t named_output'''
    clear_now_and_after('t_no.txt')
    execute_workflow(
        '''
        [A]
        output: res='t_no.txt'
        _output.touch()
        ''', targets=['res'])
    assert os.path.isfile('t_no.txt')


def test_provides_target(clear_now_and_after):
    '''Test sos run -t filename with exact match'''
    clear_now_and_after('t_pa.txt')
    execute_workflow(
        '''
        [A: provides="t_pa.txt"]
        _output.touch()
        ''', targets=['t_pa.txt'])
    assert os.path.isfile('t_pa.txt')
    #
    clear_now_and_after('t_pa.txt')
    execute_workflow(
        '''
        [A: provides="t_pa.txt"]
        output: 't_pa.txt'
        _output.touch()
        ''',
        targets=['t_pa.txt'])
    assert os.path.isfile('t_pa.txt')
    #
    clear_now_and_after('t_pa.txt')
    execute_workflow(
        '''
        [A: provides="t_pa.txt"]
        output: pa='t_pa.txt'
        _output.touch()
        ''',
        targets=['t_pa.txt'])
    assert os.path.isfile('t_pa.txt')
    #
    clear_now_and_after('t_pa.txt')
    with pytest.raises(Exception):
        execute_workflow(
            '''
            [A: provides="t_pa.txt"]
            output: pa='t_pa_none.txt'
            _output.touch()
            ''',
            targets=['t_pa.txt'])


def test_provides_pattern(clear_now_and_after):
    '''Test sos run -t filename with pattern matching'''
    clear_now_and_after('t_ma.txt')
    execute_workflow('''
        [A: provides="{filename}.txt"]
        _output.touch()
        ''', targets=['t_ma.txt'])
    assert os.path.isfile('t_ma.txt')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import argparse
import json
import operator
import random
import string
from collections import defaultdict
from itertools import tee

import numpy
import pandas


def is_sorted(iterable, compare=operator.le):
    a, b = tee(iterable)
    next(b, None)
    return all(map(compare, a, b))


class Visualizer:

    def __init__(self, kernel, style):
        self.kernel = kernel
        if style is None:
            self.style = "table"
            self.options = []
        else:
            self.style = "table" if style["style"] is None else style["style"]
            self.options = style["options"]

    def _parse_error(self, msg):
        if self.kernel:
            self.kernel.warn(msg)

    def preview(self, df):
        if self.style == "table":
            return self._handle_table(df)
        if self.style == "scatterplot":
            return self._handle_scatterplot(df)
        raise ValueError(f"Unknown style {self.style}")

    def get_tid(self, vis_type):
        alphabet = string.ascii_lowercase + string.digits
        return "".join(random.choices(alphabet, k=8))

    #
    # TABLE
    #
    def _get_table_parser(self):
        parser = argparse.ArgumentParser(
            prog="%preview -s table",
            description="""Preview Pandas DataFrame (or csv files) in tablular
            format with search and sort capacity""",
        )
        parser.add_argument(
            "-l",
            "--limit",
            type=int,
            default=200,
            help="""Limit the number of displayed records. A negative value indicates
            showing all records.""",
        )
        parser.error = self._parse_error
        return parser

    def _is_numeric_type(self, x):
        try:
            return numpy.issubdtype(x, numpy.number)
        except Exception:
            # this includes Pandas category type
            return False

    def _handle_table(self, df):
        parser = self._get_table_parser()
        try:
            args = parser.parse_args(self.options)
        except SystemExit:
            return

        if not isinstance(df, pandas.core.frame.DataFrame):
            raise ValueError("Not of DataFrame type")

        tid = self.get_tid("table")

        hint = ""
        if args.limit >= 0 and df.shape[0] > args.limit and args.limit == 200:
            hint += f'<div class="sos_hint">Only the first {args.limit} of the {df.shape[0]} records are previewed. Use option --limit to set a new limit.</div><br>'
        if args.limit >= 0:
            code = (
                df.head(args.limit).to_html(index=True).replace(
                    'class="', f'id="dataframe_{tid}" class="sos_dataframe ',
                    1))
        else:
            code = df.to_html(index=True).replace(
                'class="', f'id="dataframe_{tid}" class="sos_dataframe ', 1)

        hr, rest = code.split("</tr>", 1)
        index_type = ("numeric" if isinstance(df.index, pandas.RangeIndex) else
                      "alphabetic")
        col_type = [
            "numeric" if self._is_numeric_type(x) else "alphabetic"
            for x in df.dtypes
        ]
        # pylint: disable=consider-using-f-string
        code = ("".join(
            """{} &nbsp; <i class="fa fa-sort" style="color:lightgray" onclick="sortDataFrame('{}', {}, '{}')"></th>"""
            .format(x, tid, idx, index_type if idx ==
                    0 else col_type[idx - 1]) if "<th" in x else x
            for idx, x in enumerate(hr.split("</th>"))) + "</tr>" + rest)


        # we put max-height 400px here because the notebook could be exported without using sos template
        # and associated css, resulting in very long table.
        code = (f"""
    <div class='dataframe_container' style="max-height:400px">
    <input type="text" class='dataframe_input' id="search_{tid}" """ +
                f"""onkeyup="filterDataFrame('{tid}""" +
                """')" placeholder="Search for names..">
    """ + code + """</div>""")
        return {"text/html": hint + code}

    #
    # SCATTERPLOT
    #
    def _get_scatterplot_parser(self):
        parser = argparse.ArgumentParser(prog="%preview -s scatterplot")
        parser.add_argument(
            "cols",
            nargs="*",
            help="""Columns to plot, which should all be numeric. If one
            column is specified, it is assumed to be a x-y plot with x being 0, 1, 2, 3, .... If two or
            more columns (n) are specified, n-1 series will be plotted with the first column being the
            x axis, in which case an "_index" name can be used to specify 0, 1, 2, 3, .... This option can be
            igured if the dataframe has only one or two columns.""",
        )
        parser.add_argument("--ylim", nargs=2, help="""Range of y-axis""")
        parser.add_argument("--xlim", nargs=2, help="""Range of x-axis""")
        parser.add_argument(
            "--log",
            choices=["x", "y", "xy", "yx"],
            help="""Make x-axis, y-axis, or both to logarithmic""",
        )
        parser.add_argument(
            "--width", default="50vw", help="""Width of the plot.""")
        parser.add_argument(
            "--height", default="38vw", help="""Height of the plot.""")
        parser.add_argument(
            "-b",
            "--by",
            nargs="+",
            help="""columns by which the data points are stratified.""",
        )
        parser.add_argument(
            "--show",
            nargs="+",
            help="""What to show in the plot,
            can be 'lines', 'points' or both. Default to points, and lines if x-axis is
            sorted.""",
        )
        parser.add_argument(
            "-t",
            "--tooltip",
            nargs="*",
            help="""Fields to be shown in tooltip, in addition to
            the row index and point values that would be shown by default.""",
        )
        parser.add_argument(
            "-l",
            "--limit",
            default=2000,
            help="""Maximum number
            of records to plot.""",
        )
        parser.error = self._parse_error
        return parser

    def _to_list(self, arr):
        if "int" in arr.dtype.name:
            return [int(x) for x in arr]
        return [float(x) for x in arr]

    def _natural_ticks(self, rg):
        # given a range, get natural ticks such as 0.1, 1, 10, 100, ...
        import math

        # get integer by 1 or 2 (1, 10) or (1, 100)
        logl = math.floor(math.log10(rg[0]))
        logh = math.ceil(math.log10(rg[1]))
        # small scale, let flop decide
        if logh - logl < 3:
            return None
        return list(10**x for x in range(logl, logh + 1))

    def _handle_scatterplot(self, df):
        parser = self._get_scatterplot_parser()
        try:
            args = parser.parse_args(self.options)
        except SystemExit:
            return

        if not isinstance(df, pandas.core.frame.DataFrame):
            raise ValueError("Not of DataFrame type")

        tid = str(self.get_tid("scatterplot"))

        hint: str = ""
        if df.shape[0] > args.limit:
            hint += f'<div class="sos_hint">Only the first {args.limit} of the {df.shape[0]} records are plotted. Use option --limit to set a new limit.</div><br>'

        # replacing ' ' with &nbsp and '-' with unicode hyphen will disallow webpage to separate words
        # into lines
        indexes = [
            str(x).replace(" ", "&nbsp;").replace("-", "&#8209;")
            for x in df.index
        ]

        data = df.head(args.limit)
        nrow = data.shape[0]

        if not args.cols:
            if df.shape[1] == 1:
                args.cols = ["_index", df.columns[0]]
            elif df.shape[1] == 2:
                args.cols = list(df.columns)
            else:
                raise ValueError(
                    f'Please specify columns for plot. Available columns are {" ".join(df.columns)}'
                )

        if len(args.cols) == 1:
            args.cols = ["_index", args.cols[0]]

        if args.cols[0] == "_index":
            args.tooltip = [args.cols[0]] + (
                args.tooltip if args.tooltip else [])
        else:
            args.tooltip = ["_index", args.cols[0]] + (
                args.tooltip if args.tooltip else [])

        # check datatype
        for col in args.cols + args.tooltip + (args.by if args.by else []):
            if col == "_index":
                continue
            if col not in data.columns:
                raise ValueError(f"Invalid column name {col}")

        # tooltip and --by columns does not have to be numeric
        for col in args.cols:
            if col == "_index":
                continue
            if not self._is_numeric_type(data[col].dtype):
                raise ValueError(f"Column {col} is not of numeric type")

        if args.cols[0] == "_index":
            val_x = list(range(0, nrow))
        else:
            val_x = self._to_list(data[args.cols[0]])

        all_series = []

        if args.by:
            # create seris with _by
            vals = []
            for by_col in args.by:
                vals.append(sorted(list(set(data[by_col]))))
            # outer product
            import itertools

            categories = list(itertools.product(*vals))

        for col in args.cols[1:]:
            series = {}
            series["clickable"] = True
            series["hoverable"] = True

            if col == "_index":
                val_y = list(range(1, nrow + 1))
            else:
                val_y = self._to_list(data[col])

            tooltip = [
                "<br>".join([
                    f'{"index" if t == "_index" else t}: {idxvalue if t == "_index" else df[t][idx]}'
                    for t in args.tooltip
                ])
                for idx, idxvalue in enumerate(indexes)
            ]

            all_data = [(x, y, z) for x, y, z in zip(val_x, val_y, tooltip)]

            if args.by:
                for cat in categories:
                    series = {}
                    series["label"] = (
                        col + " (" +
                        " ".join(f"{x}={y}" for x, y in zip(args.by, cat)) +
                        ")")
                    # find index of values that falls into the category
                    series["data"] = [
                        all_data[i]
                        for i in range(len(all_data))
                        if all(data[b][i] == v for b, v in zip(args.by, cat))
                    ]
                    if len(series["data"]) > 0:
                        all_series.append(series)
            else:
                series["label"] = col
                series["data"] = all_data
                all_series.append(series)

        options = defaultdict(dict)
        options["xaxis"] = {}
        options["yaxis"] = {}
        options["series"]["lines"] = {
            "show":
                is_sorted(val_x) and not args.by
                if not args.show or "lines" in args.show else False
        }
        options["series"]["points"] = {
            "show": True if not args.show or "points" in args.show else False
        }
        options["grid"]["hoverable"] = True
        options["grid"]["clickable"] = True

        # if there are actual indexes... and plot by x
        class_name = "scatterplot"
        if args.cols[0] == "_index" and not isinstance(df.index,
                                                       pandas.RangeIndex):
            options["xaxis"]["ticks"] = [
                [x, str(y)] for x, y in enumerate(indexes)
            ]
            class_name = "scatterplot_by_rowname"

        if args.xlim:
            options["xaxis"]["min"] = args.xlim[0]
            options["xaxis"]["max"] = args.xlim[1]
        if args.ylim:
            options["yaxis"]["min"] = args.ylim[0]
            options["yaxis"]["max"] = args.ylim[1]

        optfunc = ""
        if args.log and "x" in args.log:
            range_x = [min(val_x), min(val_x)]
            optfunc = """
                options['xaxis']['transform'] = function(v) { return Math.log(v); }
                options['xaxis']['inverseTransform'] = function(v) { return Math.exp(v); }
            """
            ticks = self._natural_ticks(range_x)
            if ticks:
                optfunc += f"""
                    options['xaxis']['ticks'] = {ticks!r};
                    """
            if not args.xlim:
                options["xaxis"]["min"] = range_x[0]
                options["xaxis"]["max"] = range_x[1]
        if args.log and "y" in args.log:
            range_y = [
                min([
                    min([x[1] for x in series["data"]]) for series in all_series
                ]),
                max([
                    max([x[1] for x in series["data"]]) for series in all_series
                ]),
            ]
            optfunc += """
                options['yaxis']['transform'] = function(v) { return Math.log(v); };
                options['yaxis']['inverseTransform'] = function(v) { return Math.exp(v); };
            """
            ticks = self._natural_ticks(range_y)
            if ticks:
                optfunc += f"""
                    options['yaxis']['ticks'] = {ticks!r};
                    """
            # flot does not seems to scale correctly without min/max
            if not args.ylim:
                options["yaxis"]["min"] = range_y[0]
                options["yaxis"]["max"] = range_y[1]
        code = ("""
<div class='scatterplot_container'>
<div class='""" + class_name + """' id='dataframe_scatterplot_""" + tid +
                """' width='""" + args.width + """' height='""" + args.height +
                """'></div>
<script src="https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js"></script>
<script language="javascript" type="text/javascript" src="https://www.flotcharts.org/flot/jquery.flot.js"></script>
<script>
    var options = """ + json.dumps(options) + """;""" + optfunc + """
    function plotScatterPlot""" + tid + """() {
        plot = $.plot('#dataframe_scatterplot_""" + tid + """', """ +
                json.dumps(all_series) + """, options)

    if ($('#dftooltip').length == 0) {
        $("<div id='dftooltip'></div>").css({
            position: "absolute",
            display: "none",
            border: "1px solid #fdd",
            padding: "2px",
            "background-color": "#fee",
            "z-index": 1000,
            opacity: 0.80
            }).appendTo("body");
    }
    $('#dataframe_scatterplot_""" + tid +
                """').bind("plothover", function (event, pos, item) {
            if (item) {
                $("#dftooltip").html((item.series.label + ": " +
                    item.series.data[item.dataIndex][1].toString() + "<br>" +
                    item.series.data[item.dataIndex][2]).trim()).css({top: item.pageY+5, left: item.pageX+5})
                    .fadeIn(200);
            } else {
                $("#dftooltip").hide();
            }
    });
    }
""" + """
// we will wait for the div to be displayed on the HTML/Jupyter side before we plot
// the figure. This might not be necessary but at least this is a chance for us
// to resize the div and avoid some flot error

var dt = 100;
// the frontend might be notified before the table is inserted as results.
function showFigure""" + tid + """() {
    if ( $('#dataframe_scatterplot_""" + tid +
                """').length === 0 || $.plot === undefined ) {
          dt = dt * 1.5; // slow-down checks for datatable as time goes on;
          setTimeout(showFigure""" + tid + """, dt);
          return;
   } else {
        $('#dataframe_scatterplot_""" + tid + """').css('width', '""" +
                args.width + """').css('height', '""" + args.height + """');
        plotScatterPlot""" + tid + """();
    }
}
showFigure""" + tid + """()
</script>
</div>""")
        return {"text/html": hint + code}



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import subprocess

import pytest
from sos import execute_workflow
from sos.targets import file_target

has_docker = True
try:
    subprocess.check_output("docker ps | grep test_sos", shell=True).decode()
except subprocess.CalledProcessError:
    subprocess.call("sh build_test_docker.sh", shell=True)
    try:
        subprocess.check_output(
            "docker ps | grep test_sos", shell=True).decode()
    except subprocess.CalledProcessError:
        print("Failed to set up a docker machine with sos")
        has_docker = False


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_to_host_option(clear_now_and_after):
    """Test from_remote option"""
    clear_now_and_after("to_host_testfile.txt", "to_host_linecount.txt")
    with open("to_host_testfile.txt", "w") as tf:
        for i in range(100):
            tf.write(f"line {i+1}\n")
    execute_workflow(
        """
        [10]
        output: 'to_host_linecount.txt'
        task: to_host='to_host_testfile.txt'
        sh:
            wc -l to_host_testfile.txt > to_host_linecount.txt
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert os.path.isfile("to_host_linecount.txt")
    with open("to_host_linecount.txt") as lc:
        assert lc.read().strip().startswith("100")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_to_host_option_with_named_path(clear_now_and_after):
    """Test from_remote option"""
    clear_now_and_after(
        os.path.expanduser("~/to_host_named_testfile.txt"),
        "to_host_named_linecount.txt",
    )
    with open(os.path.expanduser("~/to_host_named_testfile.txt"), "w") as tf:
        for i in range(200):
            tf.write(f"line {i+1}\n")
    execute_workflow(
        """
        [10]
        output: 'to_host_named_linecount.txt'
        task: to_host='#home/to_host_named_testfile.txt'
        sh:
            wc -l ~/to_host_named_testfile.txt > to_host_named_linecount.txt
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert os.path.isfile("to_host_named_linecount.txt")
    with open("to_host_named_linecount.txt") as lc:
        assert lc.read().strip().startswith("200")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_from_host_option(clear_now_and_after):
    """Test from_remote option"""
    clear_now_and_after("llp")
    execute_workflow(
        """
        [10]
        task: from_host='llp'
        with open('llp', 'w') as llp:
            llp.write("LLP")
        """,
        options={
            "config_file": "~/docker.yml",
            "wait_for_task": True,
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert os.path.isfile("llp")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_local_from_host_option(clear_now_and_after):
    """Test from_remote option"""
    clear_now_and_after("llp")
    execute_workflow(
        """
        [10]
        task: from_host='llp'
        sh:
        echo "LLP" > llp
        """,
        options={
            "config_file": "~/docker.yml",
            # do not wait for jobs
            "wait_for_task": True,
            "sig_mode": "force",
            "default_queue": "localhost",
        },
    )
    assert os.path.isfile("llp")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_local_from_host_option_with_named_path(clear_now_and_after):
    """Test from_remote option"""
    clear_now_and_after(os.path.expanduser("~/llp"))
    execute_workflow(
        """
        [10]
        task: from_host='#home/llp'
        sh:
        echo "LLP" > ~/llp
        """,
        options={
            "config_file": "~/docker.yml",
            # do not wait for jobs
            "wait_for_task": True,
            "sig_mode": "force",
            "default_queue": "localhost",
        },
    )
    assert os.path.isfile(os.path.expanduser("~/llp"))


def test_worker_procs():
    # test -j option
    execute_workflow(
        """
        [1]
        input: for_each=dict(i=range(10))

        bash: expand=True
        echo {i}
        """,
        options={
            "sig_mode": "force",
            "worker_proces": ["1", "localhost:2"]
        },
    )


def test_worker_procs_with_task():
    # test -j option
    execute_workflow(
        """
        [1]
        input: for_each=dict(i=range(10))

        task: trunk_size = 0

        bash: expand=True
        echo {i}
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost",
            "worker_proces": ["1", "localhost:2"],
        },
    )


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_execute(clear_now_and_after, script_factory):
    clear_now_and_after("result_remote.txt", "result_remote1.txt", "local.txt")

    test_remote_sos = script_factory("""
        [10]
        input: 'local.txt'
        output: 'result_remote.txt'
        task:

        run:
            cp local.txt result_remote.txt
            echo 'adf' >> 'result_remote.txt'
        """)
    with open("local.txt", "w") as w:
        w.write("something")

    assert 0 == subprocess.call(
        "sos remote push docker --files local.txt -c ~/docker.yml", shell=True)

    assert 0 == subprocess.call(
        f"sos run {test_remote_sos} -c ~/docker.yml -r docker -s force",
        shell=True,
    )

    assert not file_target("result_remote.txt").target_exists()

    # self.assertEqual(subprocess.call('sos preview result_remote.txt -c ~/docker.yml -r docker', shell=True), 0)
    # self.assertNotEqual(subprocess.call('sos preview result_remote.txt', shell=True), 0)
    assert 0 == subprocess.call(
        "sos remote pull docker --files result_remote.txt -c ~/docker.yml",
        shell=True)

    assert file_target("result_remote.txt").target_exists()

    # self.assertEqual(subprocess.call('sos preview result_remote.txt', shell=True), 0)
    with open("result_remote.txt") as w:
        content = w.read()
        assert "something" in content
        assert "adf" in content

    # test sos remote run
    assert 0 == subprocess.call(
        "sos remote run docker -c  ~/docker.yml --cmd cp result_remote.txt result_remote1.txt ",
        shell=True,
    )
    assert 0 == subprocess.call(
        "sos remote pull docker --files result_remote1.txt -c ~/docker.yml",
        shell=True)

    assert file_target("result_remote1.txt").target_exists()


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_workflow_remote_queue(script_factory):
    test_r_q = script_factory('''
        input: for_each=dict(i=range(2))
        output: f'test_r_q_{i}.txt'

        task: walltime='10m', cores=1, mem='1G'
        sh: expand=True
            echo `pwd` > {_output}
            echo I am {i} >> {_output}
        ''')
    assert 0 == subprocess.call(
        f"sos run {test_r_q} -c ~/docker.yml -r ts -q ts", shell=True)


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_signature_of_remote_target(clear_now_and_after, monkeypatch):
    """Test remote() target"""
    monkeypatch.setenv("SOS_DEBUG", "TASK,-")
    clear_now_and_after("remote_file.txt", "result.txt")

    with open("remote_file.txt", "w") as rf:
        rf.write("""line1
        line2
        line3
        """)
    assert 0 == subprocess.call(
        "sos remote push docker --files remote_file.txt -c ~/docker.yml",
        shell=True)
    os.remove("remote_file.txt")
    #
    wf = """
        input: remote('remote_file.txt')
        output: 'result.txt'

        task:
        sh: expand=True
            wc -l {_input} > {_output}
        """
    execute_workflow(
        wf,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
        },
    )
    assert file_target("result.txt").target_exists()
    assert open("result.txt").read().strip().startswith("3")
    #
    # now change the remote file
    with open("remote_file.txt", "w") as rf:
        rf.write("""line1
        line2
        line3
        line4
        line5
        """)
    assert 0 == subprocess.call(
        "sos remote push docker --files remote_file.txt -c ~/docker.yml",
        shell=True)
    os.remove("remote_file.txt")
    os.remove("result.txt")
    #
    execute_workflow(
        wf,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
        },
    )
    assert file_target("result.txt").target_exists()
    assert open("result.txt").read().strip().startswith("5")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_exec(clear_now_and_after):
    clear_now_and_after("result_exec.txt")
    root_dir = "/root/build" if "TRAVIS" in os.environ else "/root"
    execute_workflow(
        """
        output: 'result_exec.txt'

        task:
        sh: expand=True
            echo Output: {_output} > {_output}
            echo PWD: `pwd`. >> {_output}
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert file_target("result_exec.txt").target_exists()
    with open(file_target("result_exec.txt")) as res:
        result = res.read()
        assert "Output: result_exec.txt" in result
        assert f"PWD: {root_dir}" in result


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_exec_named_path(clear_now_and_after):
    clear_now_and_after("result_named_path.txt")
    root_dir = "/root/build" if "TRAVIS" in os.environ else "/root"

    execute_workflow(
        """
        output: '#home/result_named_path.txt'

        task:
        sh: expand=True
            echo Output: {_output} > {_output}
            echo PWD: `pwd`. >> {_output}
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert file_target("#home/result_named_path.txt").target_exists()
    with open(file_target("#home/result_named_path.txt")) as res:
        result = res.read()
        print(result)
        assert "Output: /root/result_named_path.txt" in result
        assert f"PWD: {root_dir}" in result


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_exec_workdir_named_path(clear_now_and_after):
    clear_now_and_after(file_target("#home/wd/result_workdir_named_path.txt"))
    execute_workflow(
        """
        output: '#home/wd/result_workdir_named_path.txt'

        task: workdir='/root'
        sh: expand=True
            echo Output: {_output} > {_output}
            echo PWD: `pwd`. >> {_output}
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert file_target("#home/wd/result_workdir_named_path.txt").target_exists()
    with open(file_target("#home/wd/result_workdir_named_path.txt")) as res:
        result = res.read()
        assert "Output: /root/wd/result_workdir_named_path.txt" in result
        assert "PWD: /root." in result


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_exec_workdir_wo_named_path(clear_now_and_after):
    clear_now_and_after(file_target("result_workdir_wo_named.txt"))
    with pytest.raises(Exception):
        execute_workflow(
            """
        output: 'result_workdir_wo_named.txt'

        task: workdir='/other'
        sh: expand=True
            echo Output: {_output} > {_output}
            echo PWD: `pwd`. >> {_output}
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker",
                "sig_mode": "force",
            },
        )



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import sys
import random

import pytest
from sos.eval import accessed_vars, on_demand_options
from sos.parser import SoS_Script
from sos.pattern import expand_pattern, extract_pattern
from sos.targets import executable, file_target, sos_step, sos_targets
# these functions are normally not available but can be imported
# using their names for testing purposes
from sos.utils import (WorkflowDict, as_fstring, env, get_logger,
                       split_fstring, stable_repr, fileMD5)
from sos.workflow_executor import Base_Executor, analyze_section


def test_logger():
    '''Test logging level'''
    for verbosity in [0, 1, 2, 3, 4]:
        env.verbosity = verbosity
        get_logger().debug(
            'Verbosity {}:debug message with ``empahsized text`` in between'
            .format(env.verbosity))
        get_logger().info(
            'Verbosity {}:info message with ``empahsized text`` in between'
            .format(env.verbosity))
        get_logger().warning(
            'Verbosity {}:warning message with ``empahsized text`` in between'
            .format(env.verbosity))
        get_logger().error(
            'Verbosity {}:error message with ``empahsized text`` in between'
            .format(env.verbosity))


def test_workflow_dict():
    '''Test workflow dict with attribute access'''
    env.reset()
    d = WorkflowDict()
    d['a'] = 1
    assert d['a'] == 1
    #
    d['a'] += 1
    assert d['a'] == 2


def test_pattern_match():
    '''Test snake match's pattern match facility'''
    res = extract_pattern('{a}-{b}.txt', ['file-1.txt', 'file-ab.txt'])
    assert res['a'] == ['file', 'file']
    assert res['b'] == ['1', 'ab']
    res = extract_pattern('{a}-{b}.txt', ['file--ab--cd.txt'])
    assert res['a'] == ['file--ab-']
    assert res['b'] == ['cd']
    res = extract_pattern('{path}/{to}/{file}.txt', ['/tmp//1.txt'])
    assert res['path'] == [None]
    assert res['to'] == [None]
    assert res['file'] == [None]
    res = extract_pattern('{path}/{to}/{file}.txt', ['/tmp/test/1.txt.txt'])
    assert res['path'] == ['/tmp']
    assert res['to'] == ['test']
    assert res['file'] == ['1.txt']
    # expand_pattern
    env.sos_dict = WorkflowDict({
        'os': os,
        'a': 100,
        'b': 'file name',
        'c': ['file1', 'file2', 'file 3'],
        'd': {
            'a': 'file1',
            'b': 'file2'
        },
    })
    assert expand_pattern('{b}.txt') == ['file name.txt']
    assert expand_pattern('{c}.txt') == ['file1.txt', 'file2.txt', 'file 3.txt']
    assert expand_pattern('{a}_{c}.txt') == [
        '100_file1.txt', '100_file2.txt', '100_file 3.txt'
    ]


def test_accessed_vars():
    '''Test accessed vars of a SoS expression or statement.'''
    assert accessed_vars('''a = 1''') == set()
    assert accessed_vars('''a = b + 2.0''') == {'b'}
    assert accessed_vars('''a = "C"''') == set()
    assert accessed_vars('''a = "C" + f"{D}"''') == {'D'}
    assert accessed_vars('''a = 1 + f"{D + 20:f}" ''') == {'D'}
    assert accessed_vars(
        '''k, "a.txt", "b.txt", par=f(something) ''',
        mode='eva') == {'k', 'f', 'something'}
    # this is a complicated case because the actual variable depends on the
    # result of an expression... However, in the NO-evaluation case, this is
    # the best we can do.
    assert accessed_vars('''c + f"{D + 1}" ''') == {'c', 'D'}
    assert accessed_vars('''a, b=2, c=d ''', mode='eva') == {'a', 'd'}


def test_progress_bar():
    '''Test progress bar'''
    env.verbosity = 1
    #
    script = SoS_Script('''

[1]
[2]
[3]
[4]
[5]
''')
    wf = script.workflow()
    Base_Executor(wf).run()


def test_text_repr():
    # the " as the last character can lead to problems...
    script = SoS_Script('''
run:
echo "Hi, This is from bash"''')
    wf = script.workflow()
    Base_Executor(wf).run()
    # windows does not have #! mechanism so the python code
    # would incorrectly be executed as bat
    if sys.platform == 'win32':
        return
    for text in ('"""a"""', '"b"', r'"""\na\\nb"""', r"'''a\nb'''",
                 """ "a'\\"='" """):
        script = SoS_Script(r'''
a = 1
run: expand=True
#!/usr/bin/env python
with open('tmp.txt', 'w') as tmp:
    tmp.write({} + '{}')
k = """b"""'''.format(text, '{a}'))
        wf = script.workflow()
        Base_Executor(wf).run()
        with open('tmp.txt') as tmp:
            assert tmp.read() == eval(text) + '1'
    os.remove('tmp.txt')


def test_analyze_section():
    '''Test analysis of sections (statically)'''
    script = SoS_Script('''
g1 = 'a'
g2 = 1
parameter: p1 = 5
parameter: infiles = 'a.txt'

[A_1: shared='b']
b = p1 + 2
input:  infiles
output: None
c = 5

[A_2]
b = [1, 2, 3]
input: for_each='b'
depends: 'some.txt', executable('ls')
import time
import random

r = random.randint(1, 5)
time.sleep(r)

[A_3]
input: None
print(p1)

[A_4]
input: None
task:
python: expand=True
print(f'{output}')

[A_5]
task:
print(f'{_output}')
''')
    wf = script.workflow('A')
    Base_Executor(wf)
    for section in wf.sections:
        res = analyze_section(section)
        if section.names[0][1] == '1':
            assert res['step_input'].undetermined()
            assert res['step_depends'] == sos_targets()
            assert res['step_output'] == sos_targets()
            assert res['environ_vars'] == {'b', 'p1', 'infiles'}
            assert res['signature_vars'] == {'c'}
            assert res['changed_vars'] == {'b'}
        elif section.names[0][1] == '2':
            assert res['step_input'] == sos_targets()
            assert res['step_depends'] == sos_targets('some.txt',
                                                      executable('ls'))
            assert res['step_output'].unspecified()
            # for_each will not be used for DAG
            assert res['environ_vars'] == {'b', 'for_each', 'executable'}
            assert res['signature_vars'] == {'r', 'time', 'random'}
            assert res['changed_vars'] == set()
        elif section.names[0][1] == '4':
            assert 'output' in res['signature_vars']
        elif section.names[0][1] == '5':
            assert 'output' not in res['signature_vars']


def test_on_demand_options(reset_env):
    '''Test options that are evaluated upon request.'''
    options = on_demand_options({'a': '"est"', 'b': 'c', 'c': 'e + 2'})
    env.sos_dict = WorkflowDict({
        'e': 10,
    })
    assert options['a'] == 'est'
    with pytest.raises(KeyError):
        options.__getitem__('d')
    with pytest.raises(ValueError):
        options.__getitem__('b')
    assert options['c'] == 12
    #
    env.sos_dict.set('e', 20)
    assert options['c'] == 22
    env.sos_dict.set('c', 200)
    assert options['b'] == 200


def test_stable_repr():
    assert stable_repr({1, 2, '3', '1'}) == "{'1', '3', 1, 2}"
    assert stable_repr({1: 2, 3: 4}) == "{1:2, 3:4}"
    assert stable_repr([1, 3, 4]) == "[1, 3, 4]"


def test_split_fstring():
    '''Test function to split f-string in pieces '''
    for string, pieces in [
        ('hello world', ['hello world']),
        ('hello world {', None),
        ('{ hello world', None),
        ('hello world {}', None),
        ('hello world {a b}', None),
        ('{{hello world', ['{{hello world']),
        ('hello {{}} world', ['hello {{}} world']),
        ('hello {{ world }}', ['hello {{ world }}']),
        ('hello }} world', ['hello }} world']),
        ('hello {1} world', ['hello ', '1', ' world']),
        ('hello {a+b } }} world', ['hello ', 'a+b ', ' }} world']),
        ('hello {a+b:r} }} world', ['hello ', 'a+b:r', ' }} world']),
        ('hello {{{a+b!r} }} world', ['hello {{', 'a+b!r', ' }} world']),
        ('hello {a+b + {1,2}.pop() } }} world',
         ['hello ', 'a+b + {1,2}.pop() ', ' }} world']),
    ]:
        if pieces is None:
            with pytest.raises(SyntaxError):
                split_fstring(string)
        else:
            assert split_fstring(string) == pieces


def test_as_fstring():
    '''Test as fstring '''
    for string, fstring in [
        ('hello world', 'fr"""hello world"""'),
        ('{{hello world', 'fr"""{{hello world"""'),
        ('hello {{}} world', 'fr"""hello {{}} world"""'),
        ('hello {{ world }}', 'fr"""hello {{ world }}"""'),
        ('hello }} \n world', 'fr"""hello }} \n world"""'),
        (r'hello }} \n world', 'fr"""hello }} \\n world"""'),
        ('hello {1} world', 'fr"""hello {1} world"""'),
        ('hello {a+b } }} world', 'fr"""hello {a+b } }} world"""'),
        ('hello {a+b:r} }} world', 'fr"""hello {a+b:r} }} world"""'),
        ('''hello """ \'\'\' {a+b } }} world''',
         'f\'hello """ \\\'\\\'\\\' {a+b } }} world\''),
        ('hello {{{a+b!r} }} world', 'fr"""hello {{{a+b!r} }} world"""'),
        ('hello {a+b + {1,2}.pop() } }} world',
         'fr"""hello {a+b + {1,2}.pop() } }} world"""'),
        ('''hello {'a'+b !r} }} world''',
         'fr"""hello {\'a\'+b !r} }} world"""'),
        ('''hello """ \'\'\' {'a'+"b" + {"c", "D"}.pop() } }} world''',
         '\'hello """ \\\'\\\'\\\' {0} }} world\'.format(\'a\'+"b" + {"c", "D"}.pop() )'
        ),
    ]:
        assert as_fstring(string) == fstring


def test_analyze_output_from():
    '''Test extracting of from=value option from input'''
    script = SoS_Script('''
[A_1]
input:  output_from('B')

[A_2]
input: something_unknown, sos_groups(output_from(['C1', 'C2']), by=2), group_by=1
''')
    wf = script.workflow('A')
    Base_Executor(wf)
    for section in wf.sections:
        res = analyze_section(section, analysis_type='forward')
        if section.names[0][1] == 1:
            assert res['step_depends'] == sos_targets(sos_step('B'))
        if section.names[0][1] == 2:
            assert res['step_depends'] == sos_targets(
                sos_step('C1'), sos_step('C2'))


def test_file_sig(clear_now_and_after):
    '''test save and validate of file signature'''
    clear_now_and_after('test_sig.txt', 'test_sig.txt.zapped')

    with open('test_sig.txt', 'w') as ts:
        ts.write('ba')
    a = file_target('test_sig.txt')
    a.write_sig()
    assert a.validate()
    #
    a.zap()
    assert a.validate()

    with open('test_sig.txt', 'w') as ts:
        ts.write('bac')

    assert not a.validate()

@pytest.mark.parametrize('fsize', [12354, 33554432, 34605213])
def test_file_md5(fsize, temp_factory):
    '''test save and validate of file signature'''
    fname = 'test_md5.txt'
    temp_factory(fname, size = fsize)

    partial_md5, full_md5 = fileMD5(fname, sig_type='both')
    assert partial_md5 == fileMD5(fname, sig_type='partial')
    assert full_md5 == fileMD5(fname, sig_type='full')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from .targets import BaseTarget
from .utils import env, textMD5


class Py_Module(BaseTarget):
    '''A target for a Python module.'''

    LIB_STATUS_CACHE = {}

    def __init__(self, module, version=None, autoinstall=False):
        super().__init__()
        if not isinstance(module, str):
            raise ValueError('A string is expected for module name.')
        self._module = module.strip()
        self._version = version.strip() if isinstance(version, str) else version
        for opt in ('==', '>=', '>', '<=', '<', '!='):
            if opt in self._module:
                if self._version is not None:
                    raise ValueError(
                        f"Specifying 'version=' option in addition to '{module}' is not allowed"
                    )
                self._module, self._version = [
                    x.strip() for x in self._module.split(opt, 1)
                ]
                if ',' in self._version:
                    raise ValueError(
                        f'SoS does not yet support multiple version comparisons. {self._mdoule} provided'
                    )
                self._version = opt + self._version
                break
        self._autoinstall = autoinstall

    def _install(self, name, autoinstall):
        '''Check existence of Python module and install it using command
        pip install if necessary.'''
        import importlib

        import pkg_resources
        spam_spec = importlib.util.find_spec(name)
        reinstall = False
        if spam_spec is not None:
            if self._version:
                mod = importlib.__import__(name)
                if hasattr(mod, '__version__'):
                    ver = mod.__version__
                else:
                    try:
                        ver = pkg_resources.get_distribution(name).version
                    except Exception as e:
                        env.logger.debug(
                            f'Failed to get version of {name}: {e}')
                env.logger.debug(
                    f'Comparing exiting version {ver} against requested version {self._version}'
                )
                if self._version.startswith(
                        '==') and pkg_resources.parse_version(
                            ver) == pkg_resources.parse_version(
                                self._version[2:]):
                    pass
                elif self._version.startswith(
                        '<=') and pkg_resources.parse_version(
                            ver) <= pkg_resources.parse_version(
                                self._version[2:]):
                    pass
                elif self._version.startswith(
                        '<') and not self._version.startswith(
                            '<=') and pkg_resources.parse_version(
                                ver) < pkg_resources.parse_version(
                                    self._version[1:]):
                    pass
                elif self._version.startswith(
                        '>=') and pkg_resources.parse_version(
                            ver) >= pkg_resources.parse_version(
                                self._version[2:]):
                    pass
                # the case of >
                elif self._version.startswith(
                        '>') and not self._version.startswith(
                            '>=') and pkg_resources.parse_version(
                                ver) > pkg_resources.parse_version(
                                    self._version[1:]):
                    pass
                elif self._version.startswith(
                        '!=') and pkg_resources.parse_version(
                            ver) != pkg_resources.parse_version(
                                self._version[2:]):
                    pass
                elif self._version[0] not in (
                        '=', '>', '<', '!') and pkg_resources.parse_version(
                            ver) == pkg_resources.parse_version(self._version):
                    pass
                else:
                    env.logger.warning(
                        f'Version {ver} of installed {name} does not match specified version {self._version}.'
                    )
                    reinstall = True
        if spam_spec and not reinstall:
            return True
        if not autoinstall:
            return False
        # try to install it?
        import subprocess
        cmd = ['pip', 'install'] + ([] if self._version else ['-U']) + [
            self._module + (self._version if self._version else '')
            if self._autoinstall is True else self._autoinstall
        ]
        env.logger.info(
            f'Installing python module {name} with command {" ".join(cmd)}')
        ret = subprocess.call(cmd)
        if reinstall:
            import sys
            importlib.reload(sys.modules[name])
        # try to check version
        return ret == 0 and self._install(name, False)

    def target_exists(self, mode='any'):
        if (self._module, self._version) in self.LIB_STATUS_CACHE:
            return self.LIB_STATUS_CACHE[(self._module, self._version)]
        ret = self._install(self._module, self._autoinstall)
        self.LIB_STATUS_CACHE = {
            x: y
            for x, y in self.LIB_STATUS_CACHE.items()
            if x[0] != self._module
        }
        self.LIB_STATUS_CACHE[(self._module, self._version)] = ret
        return ret

    def target_name(self):
        return self._module

    def target_signature(self, mode='any'):
        # we are supposed to get signature of the module, but we cannot
        return textMD5('Python module ' + self._module)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import subprocess
from pathlib import Path


def assertExists(fdlist):
    for fd in fdlist:
        assert os.path.exists(fd), '{} does not exist'.format(fd)

def assertNonExists(fdlist):
    for fd in fdlist:
        assert not os.path.exists(fd), '{} still exists'.format(fd)

def test_setup(test_workflow):
    assertExists([
        'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
        'ut_d2/ut_d3/ut_f3'
    ])
    assertExists(
        ['t_f1', 't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2'])
    # this is the tricky part, directory containing untracked file should remain
    assertExists(['t_d1', 't_d1/ut_f4'])

def test_remove_all_tracked(test_workflow):
    '''test list files'''
    subprocess.call('sos remove . -t -y', shell=True)
    assertExists([
        'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
        'ut_d2/ut_d3/ut_f3'
    ])
    assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
    # this is the tricky part, directory containing untracked file should remain
    assertExists(['t_d1', 't_f1', 't_d1/ut_f4'])

def test_remove_specific_tracked(test_workflow):
    # note the t_f1, which is under current directory and has to be remove specifically.
    subprocess.call('sos remove t_f1 ut_f1 t_d2 ut_d2 -t -y', shell=True)
    assertExists([
        'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
        'ut_d2/ut_d3/ut_f3', 't_d1/t_f2', 't_d1', 't_d1/ut_f4'
    ])
    assertNonExists(['t_f1', 't_d2/t_d3/t_f3'])

def test_remove_all_untracked(test_workflow):
    '''test remove all untracked files'''
    subprocess.call('sos remove . -u -y', shell=True)
    assertNonExists(['ut_d1/ut_f2', 't_d1/ut_f4', 'ut_d2/ut_d3/ut_f3'])
    assertExists([
        't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_d1', 't_f1'
    ])
    # this is the tricky part, files under the current directory are not removed
    assertExists(['ut_f1'])

def test_remove_specific_untracked(test_workflow):
    # note the t_f1, which is under current directory and has to be remove specifically.
    subprocess.call(
        'sos remove t_f1 ut_f1 ut_d1/ut_f2 t_d1 -u -y', shell=True)
    assertNonExists(['ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4'])
    assertExists([
        't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_d1', 't_f1'
    ])
    assertExists(
        ['ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'])

def test_remove_by_age(test_workflow):
    '''test remove by age'''
    subprocess.call('sos remove --age=+1h -y', shell=True)
    # nothing is removed
    assertExists([
        'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_d1/t_f2', 't_d2/t_d3/t_f3',
        't_d2/t_d3', 't_d2', 't_d1', 't_f1', 'ut_d1', 'ut_d2',
        'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
    ])
    #
    subprocess.call('sos remove -t --age=-1h -y', shell=True)
    assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
    assertExists([
        'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_f1', 't_d2/t_d3', 't_d2',
        't_d1', 'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
    ])
    #
    subprocess.call('sos remove -u --age=-1h -y', shell=True)
    assertExists([
        'ut_f1', 't_f1', 't_d2/t_d3', 't_d2', 't_d1', 'ut_d1', 'ut_d2',
        'ut_d2/ut_d3'
    ])

def test_remove_by_size(test_workflow):
    '''test remove by size'''
    subprocess.call('sos remove --size=+10M -y', shell=True)
    # nothing is removed
    assertExists([
        'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_d1/t_f2', 't_d2/t_d3/t_f3',
        't_d2/t_d3', 't_d2', 't_d1', 't_f1', 'ut_d1', 'ut_d2',
        'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
    ])
    #
    subprocess.call('sos remove -t --size=-1M -y', shell=True)
    assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
    assertExists([
        'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_f1', 't_d2/t_d3', 't_d2',
        't_d1', 'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
    ])
    #
    subprocess.call('sos remove -u --size=-1M -y', shell=True)
    assertExists([
        'ut_f1', 't_f1', 't_d2/t_d3', 't_d2', 't_d1', 'ut_d1', 'ut_d2',
        'ut_d2/ut_d3'
    ])

def test_remove_all(test_workflow):
    '''Test remove all specified files'''
    subprocess.call('sos remove ut_d1 t_d1 ut_d2/ut_d3 -y', shell=True)
    assertExists(['t_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_f1'])

def test_remove_placeholders(test_workflow):
    '''Test remaining placeholder files'''
    # let us create a fake placeholder file
    os.remove('t_f1')
    os.remove('t_d1/t_f2')
    subprocess.call('sos dryrun test.sos', shell=True)
    assert not os.path.isfile('t_f1')
    assert not os.path.isfile('t_d1/t_f2')
    Path('t_f1').touch()
    Path('t_d1/t_f2').touch()
    #
    subprocess.call('sos remove -p', shell=True)
    assert not os.path.isfile('t_f1')
    assert not os.path.isfile('t_d1/t_f2')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import sys
import time

import pytest
from sos import execute_workflow
from sos.parser import SoS_Script
from sos.targets import file_target, sos_targets
from sos.utils import env
from sos.workflow_executor import Base_Executor


def _testSignature(script, steps):
    '''Test recognizing the format of SoS script'''
    #
    # only the first step
    res = execute_workflow(
        script,
        workflow=':0',
        options={
            'sig_mode': 'force',
            'default_queue': 'localhost',
        })
    assert os.path.isfile('temp/a.txt')
    assert os.path.isfile('temp/b.txt')
    ## FIXME
    # assert res['__completed__']['__step_completed__'], steps
    # FIXME
    with open('temp/a.txt') as ta:
        assert ta.read().strip() == 'a.txt'
    with open('temp/b.txt') as tb:
        assert tb.read().strip() == 'b.txt'

    res = execute_workflow(
        script,
        workflow=':0',
        options={
            'sig_mode': 'assert',
            'default_queue': 'localhost',
        })

    assert res['__completed__']['__step_completed__'] == 0
    # all of them
    res = execute_workflow(
        script, options={
            'sig_mode': 'default',
            'default_queue': 'localhost',
        })
    # now, rerun in build mode
    res = execute_workflow(
        script, options={
            'sig_mode': 'build',
            'default_queue': 'localhost',
        })

    assert res['__completed__']['__step_completed__'] == 0
    #
    assert os.path.isfile('temp/c.txt')
    assert os.path.isfile('temp/d.txt')
    with open('temp/c.txt') as tc:
        assert tc.read().strip() == 'a.txt'
    with open('temp/d.txt') as td:
        assert td.read().strip() == 'b.txt'
    assert env.sos_dict['oa'] == sos_targets('temp/c.txt', 'temp/d.txt')
    #
    # now in assert mode, the signature should be there
    res = execute_workflow(
        script, options={
            'sig_mode': 'assert',
            'default_queue': 'localhost',
        })
    assert res['__completed__']['__step_completed__'] == 0

    #
    res = execute_workflow(
        script, options={
            'sig_mode': 'default',
            'default_queue': 'localhost',
        })
    assert res['__completed__']['__step_completed__'] == 0

    #
    # change script a little bit
    res = execute_workflow(
        ' ' * (len(script) - len(script.lstrip())) + '# comment\n' + script,
        options={
            'sig_mode': 'assert',
            'default_queue': 'localhost',
        })

    assert res['__completed__']['__step_completed__'] == 0

    # add some other variable?
    #script = SoS_Script('comment = 1\n' + text)
    #wf = script.workflow()
    #env.config['sig_mode'] = 'assert'
    #self.assertRaises(Exception, Base_Executor(wf).run)


def test_signature(clear_now_and_after, clear_signatures):
    clear_now_and_after('temp/a.txt', 'temp/b.txt')
    _testSignature(
        r"""
        [*_0]
        output: 'temp/a.txt', 'temp/b.txt'
        task:
        run('''echo "a.txt" > temp/a.txt ''')
        run('''echo "b.txt" > temp/b.txt ''')

        [1: shared={'oa':'step_output'}]
        dest = ['temp/c.txt', 'temp/d.txt']
        input: group_by='single', paired_with='dest'
        output: _dest

        run(f" cp {_input} {_dest[0]} ")
        """, 2)


def test_signature1(clear_now_and_after, clear_signatures):
    clear_now_and_after('temp/a.txt', 'temp/b.txt')
    _testSignature(
        r"""
        [*_0]
        output: 'temp/a.txt', 'temp/b.txt'

        task:
        run('''echo "a.txt" > temp/a.txt ''')
        run('''echo "b.txt" > temp/b.txt ''')

        [1: shared={'oa':'step_output'}]
        dest = ['temp/c.txt', 'temp/d.txt']
        input: group_by='single', paired_with='dest'
        output: _dest

        run(f" cp {_input} {_dest[0]} ")
        """, 2)
    # script format


def test_signature2(clear_now_and_after, clear_signatures):
    clear_now_and_after('temp/a.txt', 'temp/b.txt')
    _testSignature(
        r"""
        [*_0]
        output: 'temp/a.txt', 'temp/b.txt'

        run:
        echo "a.txt" > temp/a.txt

        run:

        echo "b.txt" > temp/b.txt

        [1: shared={'oa':'step_output'}]
        dest = ['temp/c.txt', 'temp/d.txt']
        input: group_by='single', paired_with='dest'
        output: _dest

        task:
        run: expand=True
        echo cp {_input} {_dest[0]}
        cp {_input} {_dest[0]}
        """, 2)


def test_signature_with_shared_variable(clear_now_and_after, clear_signatures):
    '''Test restoration of signature from variables.'''
    clear_now_and_after('a.txt')
    # shared
    script = r"""
        [0: shared='a']
        output: 'a.txt'
        run:
            touch a.txt

        a = 5

        [1]
        print(a)
        """
    # alias should also be recovered.
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 2
    # rerun
    env.sos_dict.pop('a')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1


def test_signature_without_output(clear_now_and_after, clear_signatures):
    # signature without output file
    clear_now_and_after('temp/a.txt', 'temp/b.txt', 'temp')

    _testSignature(
        r"""
        [*_0]
        output: []

        run:
            [ -d temp ] || mkdir temp
            echo "a.txt" > temp/a.txt
            echo "b.txt" > temp/b.txt

        [1: shared={'oa':'step_output'}]
        dest = ['temp/c.txt', 'temp/d.txt']
        input: 'temp/a.txt', 'temp/b.txt', group_by='single', paired_with='dest'
        output: _dest

        run: expand=True
            cp {_input} {_dest[0]}
        """, 2)


def test_reexecution(clear_signatures):
    '''Test -f option of sos run'''
    script = '''
        [0]
        output: 'a.txt'
        run(f"touch {_output}")
        '''
    try:
        # remove existing output if exists
        file_target('a.txt').unlink()
    except Exception:
        pass
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    # now, rerun should be much faster
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0
    # rerun takes less than 1 second
    #
    # force rerun mode
    res = execute_workflow(script, options={'sig_mode': 'ignore'})
    assert res['__completed__']['__step_completed__'] == 1
    # regularly take more than 5 seconds to execute
    try:
        # remove existing output if exists
        os.remove('a.txt')
    except Exception:
        pass


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Windows executable cannot execute bash loop.')
def test_signature_after_removal_of_files(clear_signatures,
                                          clear_now_and_after):
    '''test action to_named_path'''
    clear_now_and_after('largefile.txt')
    script = r'''
        [10]

        # generate a file
        output: 'largefile.txt'

        run: expand='${ }'
        for x in {1..1000}
        do
            echo $x >> ${_output}
        done

        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    # rerun, because this is the final target, it has to be
    # re-generated
    os.remove('largefile.txt')
    res = execute_workflow(script)
    assert os.path.isfile('largefile.txt')
    assert res['__completed__']['__step_completed__'] == 1
    #
    # we discard the signature, the step would still be
    # skipped because file signature will be calculated
    # during verification
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0
    #
    # now if we touch the file, it needs to be regenerated
    with open('largefile.txt', 'a') as lf:
        lf.write('something')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Windows executable cannot be created with chmod.')
def test_removal_of_intermediate_files(clear_signatures, clear_now_and_after):
    # if we zap the file, it
    clear_now_and_after('midfile.txt', 'finalfile.txt', 'midfile.txt.zapped')

    script = r'''
        [10]

        # generate a file
        output: 'midfile.txt'

        run: expand='${ }'
        rm -f ${_output}
        for x in {1..1000}
        do
            echo $x >> ${_output}
        done

        [20]
        output: 'finalfile.txt'
        run: expand=True
        cp {_input} {_output}
        echo "MORE" >> {_output}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 2
    #
    # remove middle file, rerun
    os.remove('midfile.txt')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    assert os.path.isfile('midfile.txt')
    #
    # now if we touch the mid file, it needs to be regenerated
    with open('midfile.txt', 'a') as lf:
        lf.write('something')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    #
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0
    #
    # if we zap the mid file, it does not need to be rerun
    subprocess.call('sos remove midfile.txt --zap -y', shell=True)
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0


def test_signature_with_parameter(clear_now_and_after):
    '''Test signature'''
    clear_now_and_after('myfile.txt')
    #
    script = r'''
        parameter: gvar = 10

        [10]
        # generate a file
        output: 'myfile.txt'
        # additional comment
        run: expand=True
        echo {gvar} > {_output:q}

        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '10'
    #
    # now if we change parameter, the step should be rerun
    res = execute_workflow(script, args=['--gvar', '20'])
    assert res['__completed__']['__step_completed__'] == 1
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '20'
    #
    # do it again, signature should be effective
    res = execute_workflow(script, args=['--gvar', '20'])
    assert res['__completed__']['__step_completed__'] == 0
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '20'

    #
    script = r'''
        parameter: gvar = 10

        [10]
        # generate a file
        output: 'myfile.txt'
        # additional comment
        run: expand=True
        echo {gvar} > {_output:q}
        '''
    res = execute_workflow(script)
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '10'
    assert res['__completed__']['__step_completed__'] == 1
    #
    # now if we change parameter, the step should be rerun
    res = execute_workflow(script, args=['--gvar', '20'])
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '20'
    assert res['__completed__']['__step_completed__'] == 1
    #
    # do it again, signature should be effective
    res = execute_workflow(script, args=['--gvar', '20'])
    assert res['__completed__']['__step_completed__'] == 0
    with open('myfile.txt') as tmp:
        assert tmp.read().strip() == '20'


def test_loop_wise_signature(clear_now_and_after):
    '''Test partial signature'''
    clear_now_and_after([f'myfile_{i}.txt' for i in range(10, 12)])
    #
    script = r'''
        parameter: gvar = 10

        [10]
        tt = [gvar]
        input: for_each='tt'
        output: f"myfile_{_tt}.txt"
        run: expand=True
        echo "DO {_tt}"
        echo {_tt} > {_output:q}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    ts = os.path.getmtime('myfile_10.txt')
    #
    # now we modify the script
    script = r'''
        parameter: gvar = 10

        [10]
        tt = [gvar, gvar + 1]
        input: for_each='tt'
        output: f"myfile_{_tt}.txt"
        run: expand=True
        echo "DO {_tt}"
        echo {_tt} > {_output:q}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0.5
    # this file is not regenerated
    assert ts == os.path.getmtime('myfile_10.txt')
    ts1 = os.path.getmtime('myfile_11.txt')
    #
    # run it again, neither needs to be rerun
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0
    assert ts == os.path.getmtime('myfile_10.txt')
    assert ts1 == os.path.getmtime('myfile_11.txt')
    #
    # change again, the second one is already there.
    script = r'''
        parameter: gvar = 10

        [10]
        tt = [gvar + 1]
        input: for_each='tt'
        output: f"myfile_{_tt}.txt"
        run: expand=True
        echo "DO {_tt}"
        echo {_tt} > {_output:q}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0
    assert ts1 == os.path.getmtime('myfile_11.txt')
    #
    for t in range(10, 12):
        with open('myfile_{}.txt'.format(t)) as tmp:
            assert tmp.read().strip() == str(t)


def test_output_from_signature(temp_factory, clear_now_and_after):
    'Test restoration of output from signature' ''
    temp_factory('1.txt', '2.txt')
    clear_now_and_after('1.out', '2.out', '1.2.out', '2.3.out')
    script = '''
        parameter: K = [2,3]

        [work_1]
        input: "1.txt", "2.txt", group_by = 'single', pattern = '{name}.{ext}'
        output: expand_pattern('{_name}.out')
        run: expand=True
        touch {_output}

        [work_2]

        input: group_by = 'single', pattern = '{name}.{ext}', paired_with = ['K']
        output: expand_pattern('{_name}.{_K}.out')
        run: expand=True
        touch {_output}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 2
    # for the second run, output should be correctly constructed
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0


def test_signature_with_vars(temp_factory, clear_now_and_after):
    '''Test revaluation with variable change'''
    temp_factory('a1.out', 'a2.out')
    clear_now_and_after('b1.out', 'b2.out')
    script = '''
        parameter: DB = {'input': ['a1.out'], 'output': ['b1.out']}
        parameter: input_file = DB['input']
        parameter: output_file =  DB['output']

        [2]
        input: input_file, group_by = 1
        output: output_file[_index]
        run: expand=True
        touch {_output}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1
    ts = os.path.getmtime('b1.out')
    #
    script = '''
        parameter: DB = {'input': ['a1.out', 'a2.out'], 'output': ['b1.out', 'b2.out']}
        parameter: input_file = DB['input']
        parameter: output_file =  DB['output']

        [2]
        input: input_file, group_by = 1
        output: output_file[_index]
        run: expand=True
        touch {_output}
        '''
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 0.5
    assert ts == os.path.getmtime('b1.out')


def test_action_signature(clear_signatures, clear_now_and_after):
    '''Test action signature'''
    clear_now_and_after('test_action.txt', 'lc.txt')
    with open('test_action.txt', 'w') as ta:
        ta.write('#something\n')
    script = r'''
        [1]
        input: 'test_action.txt'
        run: output='lc.txt', expand=True, tracked='test_action.txt'
        sleep 5
        wc -l {_input[0]} > lc.txt
        '''
    execute_workflow(script)
    # the second time, should skip
    st = time.time()
    execute_workflow(script)
    # should skip
    assert time.time() - st < 5
    #
    script = r'''
        [1]
        input: 'test_action.txt'
        print('step is changed')
        run: output='lc.txt', expand=True, tracked='test_action.txt'
        sleep 5
        wc -l {_input[0]} > lc.txt
        '''
    st = time.time()
    execute_workflow(script)
    assert time.time() - st < 5

    # force
    execute_workflow(script, options={'sig_mode': 'build'})


def test_signature_with_without_task(clear_now_and_after):
    '''Test the inclusion of task would not trigger rerun'''
    script = r'''
        [1]
        output: 'aa'
        sh:
            echo aa > aa
        '''
    clear_now_and_after('aa')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 1

    script = r'''
        [1]
        output: 'aa'
        task:
        sh:
            echo aa > aa
        '''
    res = execute_workflow(script, config={'default_queue': 'localhost'})
    assert res['__completed__']['__step_completed__'] == 0


def test_signature_with_dynamic_output(clear_signatures, clear_now_and_after):
    '''Test return of output from dynamic output'''
    clear_now_and_after([f'rep_{i}' for i in range(5)])

    script = r'''
        [1: shared={'step1': 'step_output'}]
        input: for_each={'i': range(5)}, concurrent=True
        output: dynamic(f'rep_{i}/*.res')

        import random, os
        os.makedirs(f'rep_{i}', exist_ok=True)
        path(f'rep_{i}/{random.randint(0, 10000)}.res').touch()
        '''
    res = execute_workflow(script)
    files = env.sos_dict['step1']
    assert len(files) == 5
    assert res['__completed__']['__substep_completed__'] == 5
    # rerun
    res = execute_workflow(script)
    files_again = env.sos_dict['step1']
    assert files == files_again
    assert res['__completed__']['__substep_completed__'] == 0


def test_ignore_signature(clear_signatures, clear_now_and_after):
    '''Test ignore signature mode #1028 '''
    clear_now_and_after([f'out_{i}.txt' for i in range(3)])

    execute_workflow(
        r'''
        input: for_each={'i': range(3)}, concurrent=True
        output: f'out_{i}.txt'
        sh: expand=True
        touch {_output}
        ''',
        options={'sig_mode': 'ignore'})


def test_rebuid_signature(clear_signatures, clear_now_and_after):
    '''Test rebuilding signature'''
    clear_now_and_after('a.txt')
    script = r'''
        [A_1]
        output: 'a.txt'
        _output.touch()
        '''
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_completed__'] == 1
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 1
    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 1
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 1
    # if a.txt is changed, rebuild will rerun
    with open('a.txt', 'a') as atxt:
        atxt.write('aaa')
    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 1
    # rerun?
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 1
    #
    res = execute_workflow(script, options={'sig_mode': 'force'})
    assert res['__completed__']['__substep_completed__'] == 1
    #
    res = execute_workflow(script, options={'sig_mode': 'ignore'})
    assert res['__completed__']['__substep_completed__'] == 1


def test_rebuid_signature_with_substeps(clear_signatures, clear_now_and_after):
    '''Test rebuilding signature'''
    clear_now_and_after([f'a_{i}.txt' for i in range(4)])
    script = r'''
        [A_1]
        input: for_each=dict(i=range(4))
        output: f'a_{i}.txt'
        _output.touch()
        '''
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_completed__'] == 4
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 4
    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 4
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 4
    # if a.txt is changed, rebuild will rerun
    for i in range(4):
        with open(f'a_{i}.txt', 'a') as atxt:
            atxt.write('aaa')

    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 4
    # rerun?
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 4
    #

    res = execute_workflow(script, options={'sig_mode': 'force'})
    assert res['__completed__']['__substep_completed__'] == 4
    #
    res = execute_workflow(script, options={'sig_mode': 'ignore'})
    assert res['__completed__']['__substep_completed__'] == 4


def test_rebuid_signature_with_tasks(clear_signatures, clear_now_and_after):
    '''Test rebuilding signature'''
    clear_now_and_after([f'a_{i}.txt' for i in range(4)])
    script = r'''
[A_1]
input: for_each=dict(i=range(2))
output: f'a_{i}.txt'
task:
_output.touch()
'''
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_completed__'] == 2

    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 2

    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 2

    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 2
    # if a.txt is changed, rebuild will rerun
    for i in range(2):
        with open(f'a_{i}.txt', 'a') as atxt:
            atxt.write('aaa')

    res = execute_workflow(script, options={'sig_mode': 'build'})
    assert res['__completed__']['__substep_skipped__'] == 2
    # rerun?
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 2
    #
    res = execute_workflow(script, options={'sig_mode': 'force'})
    assert res['__completed__']['__substep_completed__'] == 2
    #
    res = execute_workflow(script, options={'sig_mode': 'ignore'})
    assert res['__completed__']['__substep_completed__'] == 2


def test_signature_with_dependency_tracing_and_vars(clear_signatures,
                                                    clear_now_and_after):
    '''Test signature with parameters with option -T #1200'''
    clear_now_and_after('out.txt')
    script = r'''
        [analyze]
        parameter: par=5
        output: 'out.txt'
        print(par)
        _output.touch()

        [default]
        input: 'out.txt'
        '''
    res = execute_workflow(
        script, options={'trace_existing': True}, args=['--par', '10'])
    assert res['__completed__']['__substep_completed__'] == 2
    # change parameter, the step should not be skipped
    res = res = execute_workflow(
        script, options={'trace_existing': True}, args=['--par', '20'])
    assert res['__completed__']['__substep_completed__'] == 2


def test_skip_mode(clear_signatures, temp_factory, clear_now_and_after):
    '''Test skipping mode of signature'''
    clear_now_and_after([f'a_{i}.bak' for i in range(4)])
    temp_factory([f'a_{i}.txt' for i in range(4)])
    #
    script = r'''
        [A_1]
        input: [f'a_{i}.txt' for i in range(4)], group_by=1
        output: _input.with_suffix('.bak')

        with open(_input) as ifile, open(_output, 'w') as ofile:
            ofile.write(ifile.read())
        '''
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_completed__'] == 4
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_skipped__'] == 4
    #
    res = execute_workflow(script, options={'sig_mode': 'skip'})
    assert res['__completed__']['__substep_skipped__'] == 4
    #
    # if result file is changed, skip will still skip
    for i in range(4):
        with open(f'a_{i}.bak', 'a') as bak:
            bak.write('extra')
    #
    res = execute_workflow(script, options={'sig_mode': 'skip'})
    assert res['__completed__']['__substep_skipped__'] == 4
    # now if we change to default mode, will not skip
    res = execute_workflow(script, options={'sig_mode': 'default'})
    assert res['__completed__']['__substep_completed__'] == 4


def test_signature_with_output_vars(clear_signatures, clear_now_and_after):
    '''Test persistence of _output with vars #1355'''
    clear_now_and_after('test_sig_with_vars.txt')
    script = '''
        [10]
        output: f'test_sig_with_vars.txt'
        _output.touch()
        _output.set(seed=1)

        [20]
        print(_input.seed)
            '''
    execute_workflow(script)
    # second time sshould be fine (using signature
    execute_workflow(script)
    #
    # worker...
    for i in range(2):
        if os.path.isfile(f'test_sig_with_vars_{i}.txt'):
            os.remove(f'test_sig_with_vars_{i}.txt')
    script = '''
        [10]
        input: for_each=dict(i=range(2))
        output: f'test_sig_with_vars_{i}.txt'
        _output.touch()
        _output.set(seed=1)

        [20]
        print(_input.seed)
            '''
    execute_workflow(script)
    # second time sshould be fine (using signature
    execute_workflow(script)
    # task
    # worker...
    for i in range(2):
        if os.path.isfile(f'test_sig_with_vars_{i}.txt'):
            os.remove(f'test_sig_with_vars_{i}.txt')
    script = '''
        [10]
        input: for_each=dict(i=range(2))
        output: f'test_sig_with_vars_{i}.txt'
        task: queue='localhost'
        _output.touch()
        _output.set(seed=1)

        [20]
        print(_input.seed)
            '''
    execute_workflow(script)
    # second time sshould be fine (using signature
    execute_workflow(script)


def test_signature_with_local_parameter(clear_signatures):
    '''test for #1372'''
    script = '''
        [default]
        input:
        output: f'constant-file-name.txt'
        parameter: n_mx = int
        bash: expand=True

        echo {n_mx} > constant-file-name.txt
        '''
    execute_workflow(script, args=['--n-mx', '60'])

    assert os.path.isfile('constant-file-name.txt')
    assert open('constant-file-name.txt').read().strip() == '60'

    execute_workflow(script, args=['--n-mx', '80'])
    assert open('constant-file-name.txt').read().strip() == '80'


#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import sys
import time
from contextlib import contextmanager

import pytest
from sos import execute_workflow
from sos.parser import SoS_Script
from sos.tasks import TaskFile, TaskParams
from sos.utils import env, textMD5
from sos.workflow_executor import Base_Executor

has_docker = sys.platform != "win32"
try:
    if sys.platform != "win32":
        subprocess.check_output(
            "docker ps | grep test_sos", shell=True).decode()
except subprocess.CalledProcessError:
    subprocess.call("sh build_test_docker.sh", shell=True)
    try:
        subprocess.check_output(
            "docker ps | grep test_sos", shell=True).decode()
    except subprocess.CalledProcessError:
        print("Failed to set up a docker machine with sos")
        has_docker = False


@contextmanager
def cd_new(path):
    old_dir = os.getcwd()
    if os.path.isdir(path):
        shutil.rmtree(path)
    os.makedirs(path, exist_ok=True)
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def get_tasks():
    from sos.signatures import WorkflowSignatures
    env.exec_dir = os.path.join(
        os.path.expanduser("~"), ".sos", textMD5(os.getcwd()))
    db = WorkflowSignatures()
    conn = db.conn
    #conn = sqlite3.connect(os.path.join(env.exec_dir, "workflow_signatures.db"))
    cur = conn.cursor()
    cur.execute('SELECT DISTINCT id FROM workflows WHERE entry_type = "task"')
    return [x[0] for x in cur.fetchall() if os.path.isfile(
        os.path.join(os.path.expanduser('~'), '.sos', 'tasks', x[0] + '.task'
    ))]


def test_task_file():
    """Test task file handling"""
    for ext in (".pulse", ".out", ".err", ".task", ".sh"):
        filename = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", "ffffffffffffffff" + ext)
        if os.path.isfile(filename):
            os.remove(filename)
    params = TaskParams(
        name="ffffffffffffffff",
        global_def=None,
        task="b=a",
        sos_dict={"a": 1},
        tags=["b", "a"],
    )
    a = TaskFile("ffffffffffffffff")
    a.save(params)
    assert a.tags == "a b"
    for ext in (".pulse", ".out", ".err", ".sh"):
        with open(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "tasks",
                    "ffffffffffffffff" + ext),
                "w",
        ) as fh:
            fh.write(ext)
    assert not a.has_stdout()
    assert not a.has_stderr()
    a.add_outputs()
    #
    assert a.params.sos_dict["a"] == 1
    assert a.status == "new"
    a.status = "completed"
    assert time.time() - a.last_updated < 2
    assert a.status == "completed"
    #
    # get and reset info
    info = a.info
    a.status = "running"
    assert a.status == "running"
    a.info = info
    assert a.status == "completed"
    assert a.has_stdout()
    #
    a.add_result({"ret_code": 5})
    #
    a.tags = ["ee", "d"]
    assert a.tags == "d ee"
    # a.add_tags(['kk'])
    # assert a.tags.split(), ['d', 'ee', 'kk'])
    #
    assert a.params.sos_dict["a"] == 1
    assert a.params.task == "b=a"
    #

    assert a.stdout == ".out"
    assert a.stderr == ".err"
    assert a.pulse == ".pulse"
    assert a.shell == ".sh"
    assert a.has_stdout()
    assert a.has_stderr()
    assert a.has_pulse()
    assert a.has_shell()
    #
    #
    a.reset()
    assert a.status == "new"
    assert a.stdout == ""
    assert a.stderr == ""
    assert a.signature == {}
    a.add_outputs()
    a.add_result({"ret_code": 5})
    assert a.result["ret_code"] == 5


def test_workdir():
    """Test workdir option for runtime environment"""
    import tempfile

    tdir = tempfile.mkdtemp()
    with open(os.path.join(tdir, "aaa.pp"), "w") as aaa:
        aaa.write("something")
    script = r"""
import os
[0]
task: workdir={0!r}

with open(os.path.join({1!r}, 'result.txt'), 'w') as res:
    for file in os.listdir({1!r}):
        res.write(file + '\n')
""".format(os.path.split(tdir)[0],
           os.path.split(tdir)[1])

    execute_workflow(
        script, options={
            "sig_mode": "force",
            "default_queue": "localhost",
        })
    with open(os.path.join(tdir, "result.txt")) as res:
        content = [x.strip() for x in res.readlines()]
        assert "aaa.pp" in content


def test_concurrency():
    """Test concurrency option for runtime environment"""
    execute_workflow(
        r"""
[0]

repeat = range(4)
input: for_each='repeat'

task:

import time
print('I am {}, waited {} seconds'.format(_index, _repeat + 1))
time.sleep(_repeat + 1)
print('I am {}, done'.format(_index))
""",
        options={
            "default_queue": "localhost",
            "sig_mode": "force"
        },
    )


def test_prepend_path():
    """Test prepend path"""
    import stat

    if not os.path.isdir("temp"):
        os.mkdir("temp")
    if sys.platform == "win32":
        with open(r"temp\temp_cmd.bat", "w") as tc:
            tc.write('echo "a"')
    else:
        with open("temp/temp_cmd", "w") as tc:
            tc.write('echo "a"')
        os.chmod("temp/temp_cmd", stat.S_IXUSR | stat.S_IWUSR | stat.S_IRUSR)
    #
    # use option env
    execute_workflow(
        script=r"""
import os
[1]
task: env={'PATH': 'temp' + os.pathsep + os.environ['PATH']}
run:
temp_cmd
""",
        options={
            "default_queue": "localhost",
            "sig_mode": "force"
        },
    )
    #
    #
    execute_workflow(
        script=r"""
[1]
task: prepend_path='temp'
run:
temp_cmd
""",
        options={"default_queue": "localhost"},
    )


def test_no_task():
    execute_workflow(
        r"""
        [10]
        task:
        run:
        sleep 0
        """,
        options={
            "default_queue": None,
            'sig_mode': 'force',
        })
    #
    env.config["sig_mode"] = "force"
    execute_workflow(
        r"""
        [10]
        task: queue=None
        run:
        sleep 0
        """,
        options={
            "default_queue": "localhost",
            'sig_mode': 'force',
        })


def test_passing_var_to_task(clear_now_and_after):
    """Test passing used variable to tasks"""
    clear_now_and_after([f"myfile_{i}.txt" for i in range(10, 13)])
    #
    env.config["sig_mode"] = "force"

    execute_workflow(
        r"""
        parameter: gvar = 10

        [10]
        # generate a file
        tt = range(gvar, gvar + 3)
        input: for_each='tt'
        output: f"myfile_{_tt}.txt"
        # additional comment

        # _tt should be used in task
        task:
        run: expand=True
        echo {_tt}_{_index} > {_output:q}

        """,
        config={
            "default_queue": "localhost",
            "sig_mode": "force",
            "max_jobs": 4
        })
    for t in range(10, 13):
        with open(f"myfile_{t}.txt") as tmp:
            assert tmp.read().strip() == str(t) + "_" + str(t - 10)


def test_max_jobs():
    """Test default max number of jobs"""
    execute_workflow(
        r"""
        [10]
        input: for_each=[{'a': range(2)}, {'b': range(3)}]

        task:
        run: expand=True
        echo "a = {a}, b = {b}"
        """,
        options={
            "max_running_jobs": 2,
            "default_queue": "localhost",
        })


def test_kill_and_purge():
    """Test no wait"""
    subprocess.call(["sos", "purge"])
    with open("test_purge.sos", "w") as script:
        script.write(r"""
[10]
task:
run:
echo Try to kill
sleep 20
""")
    subprocess.Popen("sos run test_purge.sos -s force -q localhost", shell=True)
    time.sleep(5)
    subprocess.call(["sos", "kill", "--all"])
    for i in range(20):
        output = subprocess.check_output(["sos", "status", "-v", "1"]).decode()
        if "killed" in output or "aborted" in output or "completed" in output:
            break
        assert i <= 10, "Task should be killed within 10 seconds, got {}".format(
            output)
        time.sleep(1)
    # test purge by status
    subprocess.call(["sos", "purge", "--status", "aborted"])
    assert "killed" not in subprocess.check_output(["sos", "status", "-v",
                                                    "3"]).decode()
    # purge by all is not tested because it is dangerous


def test_concurrent_task(clear_now_and_after):
    """Test submitting tasks from concurrent substeps"""
    clear_now_and_after([f"con_{x}.txt" for x in range(5)])
    execute_workflow(
        """
        [10]
        input: for_each={'i': range(5)}
        output: f'con_{i}.txt'

        task:
        run: expand=True
        echo {i} > {_output}
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost"
        })


def test_shared_option(clear_now_and_after):
    """Test shared option of task"""
    clear_now_and_after("a.txt", "a100.txt")
    execute_workflow(
        """
        [10: shared = 'a']
        output: 'a.txt'
        task: shared={'a': 'int(open("a.txt").read())'}
        run:
        echo 100 > a.txt

        [20]
        run: expand=True
        touch a{a}.txt
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost"
        })
    assert os.path.isfile("a100.txt")
    # sequence of var or mapping
    clear_now_and_after("a.txt", "a100.txt")

    execute_workflow(
        """
        [10: shared = ['a', 'b']]
        output: 'a.txt'
        task: shared=[{'a': 'int(open("a.txt").read())'}, 'b']
        b = 20
        run:
        echo 100 > a.txt

        [20]
        run: expand=True
        touch a{a}_{b}.txt
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost"
        })
    assert os.path.isfile("a100_20.txt")

    execute_workflow(
        """
        [10 (simulate): shared=['rng', 'step_rng']]
        input: for_each={'i': range(5)}
        task: shared='rng'
        print(f"{i}")
        import random
        rng = random.randint(1, 1000)
        """,
        options={"default_queue": "localhost"})

    var = env.sos_dict["rng"]
    assert isinstance(var, int)
    assert isinstance(env.sos_dict["step_rng"], list)
    assert env.sos_dict["step_rng"][-1] == var


def test_task_tags():
    """Test option tags of tasks"""
    import random

    tag = "tag{}".format(random.randint(1, 100000))
    with open("test_tags.sos", "w") as tt:
        tt.write("""
[10]
input: for_each={{'i': range(10)}}
task: tags='{}', trunk_size=2
sh: expand=True
echo {} {{i}}
""".format(tag, tag))
    wf = SoS_Script(filename="test_tags.sos").workflow()
    Base_Executor(
        wf,
        config={
            "sig_mode": "force",
            "script": "test_trunkworker.sos",
            "max_running_jobs": 10,
            "default_queue": "localhost",
            "worker_procs": ["4"],
        },
    ).run()
    ret = subprocess.check_output(
        "sos status -t {}".format(tag), shell=True).decode()
    assert len(ret.splitlines()) == 5, "Obtained {}".format(ret)
    # test multiple tags
    tag1 = "tag{}".format(random.randint(1, 100000))
    tag2 = "tag{}".format(random.randint(1, 100000))
    with open("test_tags.sos", "w") as tt:
        tt.write("""
[10]
input: for_each={{'i': range(2)}}
task: tags=['{}', '{}']
sh: expand=True
echo {} {{i}}
""".format(tag1, tag2, tag1))
    wf = SoS_Script(filename="test_tags.sos").workflow()
    Base_Executor(
        wf,
        config={
            "sig_mode": "force",
            "script": "test_trunkworker.sos",
            "max_running_jobs": 10,
            "worker_procs": ["4"],
            "default_queue": "localhost",
            "workflow": "default",
        },
    ).run()
    ret = subprocess.check_output(
        "sos status -t {}".format(tag2), shell=True).decode()
    assert len(ret.splitlines()) == 2, "Obtained {}".format(ret)


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_max_mem():
    """Test server restriction max_mem"""
    with pytest.raises(Exception):
        execute_workflow(
            """
            [10]
            task: mem='2G'
            print('test_max_mem')
            """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


def test_local_runtime_max_walltime():
    """Test server max_walltime option"""
    with pytest.raises(Exception):
        execute_workflow(
            """
            [10]
            task:
            import time
            time.sleep(15)
            """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_max_cores():
    """Test server restriction max_cores"""
    with pytest.raises(Exception):
        execute_workflow(
            """
            [10]
            task: cores=8
            print('test_max_cores')
            """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


@pytest.mark.skipIf(not has_docker, reason="Docker container not usable")
def test_override_max_cores():
    """Test use queue_args to override server restriction max_cores"""
    execute_workflow(
        """
        [10]
        task: cores=8
        print('test_override_max_cores')
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker_limited",
            "sig_mode": "force",
            "queue_args": {
                "cores": 1
            },
        },
    )


def test_list_hosts():
    """test list hosts using sos status -q"""
    for v in ["0", "1", "3", "4"]:
        output = subprocess.check_output(
            ["sos", "remote", "list", "-c", "~/docker.yml", "-v", v]).decode()
        assert "local_limited" in output, f"local_limited not in \n{output}\n for verbosity {v}"


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_max_walltime(purge_tasks):
    """Test server restriction max_walltime"""
    with pytest.raises(Exception):
        execute_workflow(
            """
            [10]
            task: walltime='1:00:00'
            print('test_max_walltime')
            """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


def test_purge_all_with_option():
    """Test sos purge all with options such as -s completed"""
    with cd_new("temp_c"):
        with open("test.sos", "w") as tst:
            tst.write("""
input: for_each={'i': range(2)}
output: f'a{i}.txt'
task:
run: expand=True
echo temp_a
touch {_output}
""")
        subprocess.call("sos run test -s force -q localhost", shell=True)
        tasks = get_tasks()
        subprocess.call("sos purge -s failed", shell=True)
    # check tasks
    taskstatus = [
        x.split()[0] for x in subprocess.check_output(
            "sos status -v1", shell=True).decode().splitlines()
    ]
    assert all(x in taskstatus for x in tasks)
    # purge one of them
    subprocess.call(f"sos purge {tasks[0]}", shell=True)
    taskstatus = [
        x.split()[0] for x in subprocess.check_output(
            "sos status -v1", shell=True).decode().splitlines()
    ]
    assert tasks[0] not in taskstatus
    assert tasks[1] in taskstatus
    #
    subprocess.call("sos purge --all", shell=True)
    taskstatus = [
        x.split()[0] for x in subprocess.check_output(
            "sos status -v1", shell=True).decode().splitlines()
    ]
    assert tasks[1] not in taskstatus


def test_resubmit_task_with_different_walltime():
    """Test resubmission of tasks with different walltime #1019"""
    with cd_new("temp_walltime"):
        with open("test.sos", "w") as tst:
            tst.write("""
task: walltime='1m'
sh:
echo 0.1
""")
        subprocess.call("sos run test -s force -q localhost", shell=True)
        tasks = get_tasks()
        out = subprocess.check_output(f"sos status {tasks[0]} -v4", shell=True)
        assert "00:01:00" in out.decode()
        with open("test1.sos", "w") as tst:
            tst.write("""
task: walltime='2m'
sh:
echo 0.1
""")
        subprocess.call("sos run test1 -s force -q localhost", shell=True)
        new_tasks = get_tasks()
        assert tasks == new_tasks
        #
        out = subprocess.check_output(f"sos status {tasks[0]} -v4", shell=True)
        assert "00:02:00" in out.decode()


def test_task_signature(purge_tasks):
    """Test re-execution of tasks"""
    with cd_new("temp_signature"):
        with open("test.sos", "w") as tst:
            tst.write("""
task:
sh:
    sleep 2
""")
        subprocess.call("sos run test -s force -q localhost", shell=True)
        tasks = get_tasks()
        tf = TaskFile(tasks[0])
        assert tf.has_signature()
        assert tf.status == "completed"
        tf.tags_created_start_and_duration()
        #
        subprocess.call("sos run test -q localhost", shell=True)
        assert tf.tags_created_start_and_duration()[3] < 1


def test_wrong_host():
    with pytest.raises(Exception):
        execute_workflow(
            """
            [10]
            task: walltime='1:00:00', queue='undefined'
            print('test_wrong_host')
            """,
            options={"default_queue": "localhost"})


def test_output_in_task():
    """Test passing _output to task #1136"""
    execute_workflow(
        """
        chunks  = [1,2]
        [1]
        input: for_each = 'chunks'
        output: f'{_chunks}.txt'
        _output.touch()

        [2]
        input: group_with = 'chunks'
        output: summary_stats = f'{_input}.summary', ld_matrix = f'{_input}.result'
        task:

        python3: expand="${ }"
            open("${_output['summary_stats']}", 'w').close()
            open("${_output['ld_matrix']}", 'w').close()
        """,
        options={"default_queue": "localhost"})


def test_repeated_tasks():
    """Test statement before task #1142 """
    for i in range(5):
        execute_workflow(
            """
            [1]
            input: for_each=dict(i=range(5))

            print(f'This is for {i}')
            task:  walltime='10m'
            print(f'this is task {i}')

            """,
            options={
                "sig_mode": "force",
                "default_queue": "localhost"
            })


def test_passing_parameters_to_task():
    """Test passing of parameters in global section to tasks #1155"""
    execute_workflow(
        """\
        [global]
        parameter: par=5
        def a():
            print(par)

        [default]
        task:
        a()
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost"
        })


def test_trunk_size_with_stop_if(clear_now_and_after):
    """Test a case when some tasks are not submitted due to holes in slots #1159"""
    clear_now_and_after([f"{i+1}.txt" for i in range(5)])
    execute_workflow(
        """\
        [1]
        output: [f'{x+1}.txt' for x in range(5)]
        for i in range(5):
            name = f'{i+1}.txt'
            if i not in [0,1,2]:
                path(name).touch()
            else:
                with open(name, 'w') as f:
                    f.write('test it')

        [2]
        input: group_by = 1
        output: f'{_input:n}.out'
        stop_if(_input.stat().st_size==0, no_output=True)

        task: trunk_size = 80
        _output.touch()
        """,
        options={
            "sig_mode": "force",
            "default_queue": "localhost"
        })


def test_output_from_master_task():
    """Test splitting the output from master task #1203"""
    execute_workflow(
        """\
        l=[x for x in range(1,13)]

        [2]
        input: for_each = 'l'
        output: f'{_l}.out'

        task: trunk_size = 4
        _output.touch()

        [3]
        assert _input == f'{_index+1}.out'
        """,
        options={"default_queue": "localhost"})


@pytest.mark.skipIf(not has_docker, reason="Docker container not usable")
def test_remote_input_target(clear_now_and_after):
    """Test the use of remote target"""
    clear_now_and_after("vars.sh", "vars1.sh")
    execute_workflow(
        """
        [10]
        input: remote('/lib/init/vars.sh')
        output: f'vars1.sh'

        task:

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert os.path.isfile("vars1.sh")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_delayed_interpolation(clear_now_and_after):
    """Test delayed interpolation with expression involving remote objects"""
    # purge all previous tasks
    clear_now_and_after('test.py', 'test.py.bak')
    execute_workflow(
        """
        [10]
        output: remote('test.py')
        task:
        run:
            touch test.py

        [20]
        output: remote(f"{_input:R}.bak")
        task:
        run: expand=True
            cp {_input} {_output}
        """,
        options={
            "config_file": "~/docker.yml",
            # do not wait for jobs
            "wait_for_task": True,
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    # this file is remote only
    assert not os.path.isfile("test.py")
    assert not os.path.isfile("test.py.bak")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target(clear_now_and_after):
    """Test the use of remote target"""
    clear_now_and_after("vars.sh", "vars1.sh")
    execute_workflow(
        """
        [10]
        input: remote('/lib/init/vars.sh')
        output: remote(f'vars1.sh')

        task:

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("vars1.sh")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target_with_trunksize(clear_now_and_after):
    clear_now_and_after("vars.sh", "vars1.sh")

    execute_workflow(
        """\
        [10]
        import os
        input: remote('/lib/init/vars.sh'), remote('/lib/init/init-d-script'), group_by=1
        output: remote(os.path.basename(str(_input)))

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())""",
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("init-d-script")


def test_runtime_max_walltime():
    """Test server max_walltime option"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task:
        import time
        time.sleep(25)
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_sync_master_task(clear_now_and_after):
    """Test sync input and output with remote host with trunksize"""
    clear_now_and_after([f"test_{i}.txt" for i in range(4)],
                        [f"test_{i}.bak" for i in range(4)])
    import random

    val = random.randint(1, 10000)
    execute_workflow(
        r"""
        parameter: g = 100

        [10]
        input: for_each=dict(i=range(4))
        output: f'test_{i}.txt'

        with open(f'test_{i}.txt', 'w') as tst:
            tst.write(f'test_{i}_{g}')

        [20]
        output: _input.with_suffix('.bak')

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read() + '.bak')
        """,
        args=["--g", str(val)],
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    # now check if
    for i in range(4):
        assert os.path.isfile(f"test_{i}.txt")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"
        assert os.path.isfile(f"test_{i}.bak")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"


def test_local_max_cores():
    """Test server restriction max_cores"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: cores=8
        print('test_local_max_cores')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )


def test_local_max_mem():
    """Test server restriction max_mem"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: mem='2G'
        print('test_local_max_mem')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )


def test_trunk_size_option(clear_now_and_after, purge_tasks):
    """Test option trunk_size"""
    clear_now_and_after([f"{i}.txt" for i in range(10)])
    execute_workflow(
        """
[10]
input: for_each={'I': range(10)}
task: trunk_size=5, walltime='10m'
run: expand=True
  echo {I} > {I}.txt
  sleep 0.1
""",
        options={
            "sig_mode": "force",
            "default_queue": "localhost",
        },
    )

    for i in range(10):
        assert os.path.isfile(f"{i}.txt")


def test_trunk_size_none_option(clear_now_and_after, purge_tasks):
    """Test option trunk_size"""
    clear_now_and_after([f"{i}.txt" for i in range(10)])
    # trunk size is None or 0, -1, intepreted as all tasks
    execute_workflow(
        """
[10]
input: for_each={'I': range(10)}
task: trunk_size=None, cores=1, walltime='10m'
run: expand=True
  echo {I} > {I}.txt
  sleep 0.1
""",
        options={
            "sig_mode": "force",
            "default_queue": "localhost",
        },
    )
    for i in range(10):
        assert os.path.isfile(f"{i}.txt")


def test_trunk_workers_option(clear_now_and_after, purge_tasks):
    """Test option trunk_workers"""
    clear_now_and_after([f"{i}.txt" for i in range(12)])

    execute_workflow(
        """
[10]
input: for_each={'I': range(12)}
task: trunk_size=6, trunk_workers=3, walltime='10m'
run: expand=True
  echo {I} > {I}.txt
  sleep 1
""",
        options={
            "max_running_jobs": 10,
            "worker_procs": ["4"],
            "default_queue": "localhost"
        },
    )
    for i in range(12):
        assert os.path.isfile(f"{i}.txt")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_sync_input_output_and_rerun(clear_now_and_after, purge_tasks):
    """Test sync input and output with remote host"""
    clear_now_and_after([f'test_{i}.txt' for i in range(4)])
    clear_now_and_after([f'test_{i}.bak' for i in range(4)])

    import random
    val = random.randint(1, 10000)
    wf = """
        parameter: g = 100

        [10]
        input: for_each=dict(i=range(4))
        output: f'test_{i}.txt'

        with open(f'test_{i}.txt', 'w') as tst:
            tst.write(f'test_{i}_{g}')

        [20]
        output: _input.with_suffix('.bak')

        task:

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read() + '.bak')
        """
    execute_workflow(
        wf,
        args=["--g", str(val)],
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )

    # now check if
    for i in range(4):
        assert os.path.isfile(f"test_{i}.txt")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"

        assert os.path.isfile(f"test_{i}.bak")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"
    #
    # test rerun the task file on local host
    clear_now_and_after([f'test_{i}.txt' for i in range(4)])
    clear_now_and_after([f'test_{i}.bak' for i in range(4)])

    execute_workflow(wf, args=["--g", str(val)], options={"sig_mode": "force"})

    for i in range(4):
        assert os.path.isfile(f"test_{i}.txt")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"

        assert os.path.isfile(f"test_{i}.bak")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"
